#!/usr/bin/env python3

# Subclass of fract4d.fractal.T which works with a GUI

import os
import struct
import math
import copy

import gi

import cairo
gi.require_foreign('cairo')  # ensure Cairo integration support is available

gi.require_version('Gdk', '3.0')
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, Gdk, GObject, GdkPixbuf, GLib

from fract4d_compiler import fracttypes, function
from fract4d import fractal, fract4dc, image, messages

from . import utils, fourway


class Hidden(GObject.GObject):
    """This class implements a fractal which calculates asynchronously
    and is integrated with the GTK main loop"""
    __gsignals__ = {
        'parameters-changed': (
            (GObject.SignalFlags.RUN_FIRST | GObject.SignalFlags.NO_RECURSE),
            None, ()),
        'iters-changed': (
            (GObject.SignalFlags.RUN_FIRST | GObject.SignalFlags.NO_RECURSE),
            None, (GObject.TYPE_INT,)),
        'tolerance-changed': (
            (GObject.SignalFlags.RUN_FIRST | GObject.SignalFlags.NO_RECURSE),
            None, (GObject.TYPE_FLOAT,)),
        'formula-changed': (
            (GObject.SignalFlags.RUN_FIRST | GObject.SignalFlags.NO_RECURSE),
            None, ()),
        'status-changed': (
            (GObject.SignalFlags.RUN_FIRST | GObject.SignalFlags.NO_RECURSE),
            None, (GObject.TYPE_INT,)),
        'progress-changed': (
            (GObject.SignalFlags.RUN_FIRST | GObject.SignalFlags.NO_RECURSE),
            None, (GObject.TYPE_FLOAT,)),
        'pointer-moved': (
            (GObject.SignalFlags.RUN_FIRST | GObject.SignalFlags.NO_RECURSE),
            None, (GObject.TYPE_INT,
                   GObject.TYPE_FLOAT, GObject.TYPE_FLOAT)),
        'stats-changed': (
            (GObject.SignalFlags.RUN_FIRST | GObject.SignalFlags.NO_RECURSE),
            None, (GObject.TYPE_PYOBJECT,))

    }

    def __init__(self, comp, width, height, total_width=-1, total_height=-1):
        GObject.GObject.__init__(self)

        (self.readfd, self.writefd) = os.pipe()
        self.nthreads = 1

        self.compiler = comp

        self.x = self.y = 0
        self.button = 0
        self.last_progress = 0.0
        self.skip_updates = False
        self.running = False
        self.frozen = False  # if true, don't emit signals

        self.site = fract4dc.fdsite_create(self.writefd)
        self.f = None
        self.try_init_fractal()

        self.input_add(self.readfd, self.onData)

        self.width = width
        self.height = height
        self.image = image.T(
            self.width, self.height, total_width, total_height)

        self.msgbuf = b""

    def try_init_fractal(self):
        f = fractal.T(self.compiler, self.site)
        self.set_fractal(f)
        self.f.compile()

    def set_fractal(self, f):
        if f != self.f:
            if self.f:
                self.interrupt()
                del self.f
            self.f = f

            # take over fractal's changed function
            f.changed = self.changed
            f.formula_changed = self.formula_changed
            f.warn = self.warn
            self.formula_changed()
            self.changed()

    def changed(self, clear_image=True):
        if self.f is None:
            return
        self.f.dirty = True
        self.f.clear_image = clear_image
        self.set_saved(False)
        if not self.frozen:
            self.emit('parameters-changed')

    def formula_changed(self):
        # print "formula changed"
        self.f.dirtyFormula = True
        # if not self.frozen:
        self.emit('formula-changed')

    def set_saved(self, val):
        if self.f is not None:
            self.f.saved = val

    def input_add(self, fd, cb):
        utils.input_add(fd, cb)

    def error(self, msg, err):
        print("Error: %s %s" % (msg, err))

    def warn(self, msg):
        print("Warning: ", msg)

    def update_formula(self):
        if self.f is not None:
            self.f.dirtyFormula = True

    def freeze(self):
        self.frozen = True

    def thaw(self):
        if self.f is None:
            return False

        self.frozen = False
        was_dirty = self.f.dirty
        self.f.clean()
        return was_dirty

    def interrupt(self):
        if self.skip_updates:
            # print "skip recursive interrupt"
            return

        self.skip_updates = True

        fract4dc.interrupt(self.site)

        n = 0
        # wait for stream from worker to flush
        while self.running:
            n += 1
            Gtk.main_iteration()

        self.skip_updates = False

    def copy_f(self):
        return copy.copy(self.f)

    def set_formula(self, fname, formula, index=0):
        self.f.set_formula(fname, formula, index)

    def onData(self, fd, condition):
        self.msgbuf = self.msgbuf + \
            os.read(fd, 8 - len(self.msgbuf))

        if len(self.msgbuf) < 8:
            # print "incomplete message: %s" % list(self.msgbuf)
            return True

        (t, size) = struct.unpack("2i", self.msgbuf)
        self.msgbuf = b""
        bytes = os.read(fd, size)
        if len(bytes) < size:
            print(
                "not enough bytes, got %d instead of %d" %
                (len(bytes), size))
            return True

        m = messages.parse(t, bytes)

        if utils.threads_enabled:
            Gdk.threads_enter()

        # print "msg: %s %d %d %d %d" % (m,p1,p2,p3,p4)
        if t == fract4dc.MESSAGE_TYPE_ITERS:
            if not self.skip_updates:
                self.iters_changed(m.iterations)
        elif t == fract4dc.MESSAGE_TYPE_IMAGE:
            if not self.skip_updates:
                self.image_changed(m.x, m.y, m.w, m.h)
        elif t == fract4dc.MESSAGE_TYPE_PROGRESS:
            if not self.skip_updates:
                progress = m.progress
                # filters out 'backwards' progress which can occur due to
                # threading
                if progress > self.last_progress or progress == 0.0:
                    self.progress_changed(progress)
                    self.last_progress = progress
        elif t == fract4dc.MESSAGE_TYPE_STATUS:
            if m.status == fract4dc.CALC_DONE:  # DONE
                self.running = False
            if not self.skip_updates:
                self.status_changed(m.status)
        elif t == fract4dc.MESSAGE_TYPE_PIXEL:
            # FIXME pixel_changed
            pass
        elif t == fract4dc.MESSAGE_TYPE_TOLERANCE:
            # tolerance changed
            if not self.skip_updates:
                self.tolerance_changed(m.tolerance)
        elif t == fract4dc.MESSAGE_TYPE_STATS:
            if not self.skip_updates:
                self.stats_changed(m)
        else:
            print("Unknown message from fractal thread; %s" % list(bytes))

        if utils.threads_enabled:
            Gdk.threads_leave()
        return True

    def __getattr__(self, name):
        return getattr(self.f, name)

    def params(self):
        return self.f.params

    def get_param(self, n):
        return self.f.get_param(n)

    def set_nthreads(self, n):
        if self.nthreads != n:
            self.nthreads = n
            self.changed()

    def set_auto_deepen(self, deepen):
        if self.f.auto_deepen != deepen:
            self.f.auto_deepen = deepen
            self.changed()

    def set_antialias(self, aa_type):
        if self.f.antialias != aa_type:
            self.f.antialias = aa_type
            self.changed()

    def set_func(self, func, fname, formula):
        self.f.set_func(func, fname, formula)

    def improve_quality(self):
        self.freeze()
        self.set_maxiter(self.f.maxiter * 2)
        self.set_period_tolerance(self.f.period_tolerance / 10.0)
        self.thaw()
        self.changed()

    def reset(self):
        self.f.reset()
        self.changed()

    def loadFctFile(self, file):
        new_f = fractal.T(self.compiler, self.site)
        new_f.warn = self.warn
        new_f.loadFctFile(file)
        self.set_fractal(new_f)
        self.set_saved(True)

    def is_saved(self):
        if self.f is None:
            return True
        return self.f.saved

    def save_image(self, filename):
        self.image.save(filename)

    def progress_changed(self, progress):
        self.emit('progress-changed', progress)

    def status_changed(self, status):
        self.emit('status-changed', status)

    def iters_changed(self, n):
        self.f.maxiter = n
        # don't emit a parameters-changed here to avoid deadlock
        self.emit('iters-changed', n)

    def tolerance_changed(self, tolerance):
        self.f.period_tolerance = tolerance
        self.emit('tolerance-changed', tolerance)

    def image_changed(self, x1, y1, x2, y2):
        pass

    def stats_changed(self, stats):
        self.emit('stats-changed', stats)

    def draw(self, image, width, height, nthreads):
        t = self.f.epsilon_tolerance(width, height)
        if self.f.auto_epsilon:
            self.f.set_named_param("@epsilon", t,
                                   self.f.formula, self.f.initparams)

        self.f.init_pfunc()
        cmap = self.f.get_colormap()
        self.running = True
        try:
            self.f.calc(image, cmap, nthreads, self.site, True)
        except MemoryError:
            pass

    def draw_image(self, aa=None, auto_deepen=None):
        if self.f is None:
            return
        self.interrupt()

        self.f.compile()

        if aa is not None and auto_deepen is not None:
            self.f.antialias = aa
            self.f.auto_deepen = auto_deepen

        self.draw(self.image, self.width, self.height, self.nthreads)

    def set_plane(self, angle1, angle2):
        self.freeze()
        self.reset_angles()
        if angle1 is not None:
            self.set_param(angle1, math.pi / 2)
        if angle2 is not None:
            self.f.set_param(angle2, math.pi / 2)

        if self.thaw():
            self.changed()

    def float_coords(self, x, y):
        return ((x - self.width / 2.0) / self.width,
                (y - self.height / 2.0) / self.width)

    def recenter(self, x, y, zoom):
        dx = (x - self.width / 2.0) / self.width
        dy = (y - self.height / 2.0) / self.width
        self.relocate(dx, dy, zoom)

    def count_colors(self, rect):
        # calculate the number of different colors which appear
        # in the subsection of the image bounded by the rectangle
        xstart, ystart, xend, yend = map(int, rect)
        buf = self.image.image_buffer(0, 0)
        colors = {}
        for y in range(ystart, yend):
            for x in range(xstart, xend):
                offset = (y * self.width + x) * 3
                col = buf[offset:offset + 3].hex()
                colors[col] = 1 + colors.get(col, 0)
        return len(colors)

    def get_func_name(self):
        if self.f is None:
            return _("No fractal loaded")
        return self.f.forms[0].funcName

    def get_saved(self):
        if self.f is None:
            return True
        return self.f.get_saved()

    def serialize(self, compress=False):
        if self.f is None:
            return None
        return self.f.serialize(compress)

    def set_size(self, new_width, new_height):
        self.interrupt()
        if self.width == new_width and self.height == new_height:
            return

        self.width = new_width
        self.height = new_height

        self.image.resize_full(new_width, new_height)
        utils.idle_add(self.changed)


# explain our existence to GTK's object system
GObject.type_register(Hidden)


class HighResolution(Hidden):
    "An invisible GtkFractal which computes in multiple chunks"

    def __init__(self, comp, width, height):
        (tile_width, tile_height) = self.compute_tile_size(width, height)

        Hidden.__init__(self, comp, tile_width, tile_height, width, height)
        self.reset_render()

    def reset_render(self):
        self.tile_list = self.image.get_tile_list()
        self.ntiles = len(self.tile_list)
        self.ncomplete_tiles = 0
        self.last_overall_progress = 0.0

    def compute_tile_size(self, w, h):
        tile_width = w
        tile_height = min(h, 128)
        return (tile_width, tile_height)

    def draw_image(self, name):
        if self.f is None:
            return
        self.interrupt()

        self.f.compile()

        self.f.auto_deepen = False
        self.f.auto_tolerance = False
        self.image.start_save(name)
        self.next_tile()
        return False

    def next_tile(self):
        # work left to do
        (xoff, yoff, w, h) = self.tile_list.pop(0)
        self.image.resize_tile(w, h)
        self.image.set_offset(xoff, yoff)
        self.draw(self.image, w, h, self.nthreads)

    def status_changed(self, status):
        if status == 0:
            # done this chunk
            self.image.save_tile()
            self.ncomplete_tiles += 1
            if len(self.tile_list) > 0:
                self.next_tile()
            else:
                # completely done
                self.image.finish_save()
                self.emit('status-changed', status)
        else:
            self.emit('status-changed', status)

    def progress_changed(self, progress):
        overall_progress = (
            100.0 * self.ncomplete_tiles + progress) / self.ntiles
        if overall_progress > self.last_overall_progress:
            self.emit('progress-changed', overall_progress)
            self.last_overall_progress = overall_progress


class T(Hidden):
    "A visible GtkFractal which responds to user input"
    SELECTION_LINE_WIDTH = 2.0

    def __init__(self, comp, parent=None, width=640, height=480):
        self.parent = parent
        Hidden.__init__(self, comp, width, height)

        self.paint_mode = False

        drawing_area = Gtk.DrawingArea()
        drawing_area.set_events(
            Gdk.EventMask.BUTTON_RELEASE_MASK |
            Gdk.EventMask.BUTTON1_MOTION_MASK |
            Gdk.EventMask.POINTER_MOTION_MASK |
            Gdk.EventMask.POINTER_MOTION_HINT_MASK |
            Gdk.EventMask.BUTTON_PRESS_MASK |
            Gdk.EventMask.KEY_PRESS_MASK |
            Gdk.EventMask.KEY_RELEASE_MASK |
            Gdk.EventMask.EXPOSURE_MASK
        )

        self.notice_mouse = False

        drawing_area.connect('motion_notify_event', self.onMotionNotify)
        drawing_area.connect('button_release_event', self.onButtonRelease)
        drawing_area.connect('button_press_event', self.onButtonPress)
        drawing_area.connect('draw', self.redraw_rect)

        self.selection_rect = []

        drawing_area.set_size_request(self.width, self.height)

        self.widget = drawing_area

    def image_changed(self, x1, y1, x2, y2):
        self.widget.queue_draw_area(x1, y1, x2 - x1, y2 - y1)

    def make_numeric_entry(self, form, param, order):
        if param.type == fracttypes.Int:
            fmt = "%d"
        else:
            fmt = "%.17f"

        widget = Gtk.Entry()
        widget.set_activates_default(True)

        def set_entry():
            new_value = fmt % form.params[order]
            if widget.get_text() != new_value:
                widget.set_text(new_value)

        def set_fractal(entry, event, form, order):
            try:
                utils.idle_add(
                    form.set_param, order, entry.get_text())
            except Exception as err:
                # FIXME: produces too many errors
                msg = "Invalid value '%s': must be a number" % \
                      entry.get_text()
                print(msg)
                # utils.idle_add(f.warn,msg)
            return False

        set_entry()

        widget.update_function = set_entry

        widget.f = self
        widget.connect('focus-out-event', set_fractal, form, order)

        if hasattr(param, "min") and hasattr(param, "max"):
            # add a slider
            adj = Gtk.Adjustment(
                value=0.0,
                lower=param.min.value, upper=param.max.value,
                step_increment=0.001, page_increment=0.01)

            def set_adj():
                if adj.get_value() != form.params[order]:
                    adj.set_value(form.params[order])

            set_adj()

            def adj_changed(adjustment, form, order):
                utils.idle_add(
                    form.set_param, order, adjustment.get_value())

            adj.connect('value-changed', adj_changed, form, order)

            hscale = Gtk.Scale.new(Gtk.Orientation.HORIZONTAL, adj)
            hscale.set_draw_value(False)
            hscale.update_function = set_adj
            vbox = Gtk.VBox()
            vbox.pack_start(widget, True, True, 0)
            vbox.pack_start(hscale, True, True, 0)
            return vbox

        return widget

    def make_numeric_widget(
            self, table, i, form, name, part, param, order):

        label = Gtk.Label.new(self.param_display_name(name, param) + part)
        label.set_halign(Gtk.Align.END)
        label.set_valign(Gtk.Align.CENTER)
        table.attach(label, 0, i, 1, 1)

        widget = self.make_numeric_entry(form, param, order)

        label.set_mnemonic_widget(widget)
        return widget

    def make_bool_widget(self, form, name, param, order):

        widget = Gtk.CheckButton(label=self.param_display_name(name, param))

        def set_toggle(*args):
            is_set = form.params[order]
            widget.set_active(is_set)
            if widget.get_active() != is_set:
                widget.set_active(is_set)

        def set_fractal(entry, form, order):
            try:
                utils.idle_add(form.set_param, order, entry.get_active())
            except Exception as err:
                msg = "error setting bool param: %s" % str(err)
                print(msg)
                #utils.idle_add(f.warn, msg)

            return False

        set_toggle(self)

        widget.update_function = set_toggle
        widget.f = self
        widget.connect('toggled', set_fractal, form, order)
        return widget

    def make_color_widget(
            self, table, i, form, name, param, order):

        label = Gtk.Label.new(self.param_display_name(name, param))
        label.set_halign(Gtk.Align.END)
        label.set_valign(Gtk.Align.CENTER)
        table.attach(label, 0, i, 1, 1)

        def set_fractal(r, g, b, is_left):
            self.freeze()
            form.set_param(order, r)
            form.set_param(order + 1, g)
            form.set_param(order + 2, b)
            if self.thaw():
                self.changed()

        rgba = []
        for j in range(4):
            rgba.append(form.params[order + j])

        color_button = utils.ColorButton(rgba, set_fractal, False)

        def set_selected_value(*args):
            rgba = []
            for j in range(4):
                rgba.append(form.params[order + j])
            color_button.set_color(rgba)

        set_selected_value()

        color_button.update_function = set_selected_value

        return color_button

    def make_enumerated_widget(
            self, table, i, form, name, part, param, order):

        label = Gtk.Label.new(self.param_display_name(name, param))
        label.set_halign(Gtk.Align.END)
        label.set_valign(Gtk.Align.CENTER)
        table.attach(label, 0, i, 1, 1)

        widget = utils.create_option_menu(param.enum.value)

        def set_selected_value(*args):
            try:
                index = form.params[order]
            except ValueError as err:
                print(err)
                return

            utils.set_selected(widget, index)

        def set_fractal(entry, form, order):
            new_value = utils.get_selected(widget)
            form.set_param(order, new_value)

        set_selected_value(self)

        widget.update_function = set_selected_value

        widget.f = self
        widget.connect('changed',
                       set_fractal, form, order)

        label.set_mnemonic_widget(widget)
        return widget

    def add_formula_setting(
            self, table, i, form, name, part, param, order):

        if param.type == fracttypes.Int:
            if hasattr(param, "enum"):
                widget = self.make_enumerated_widget(
                    table, i, form, name, part, param, order)
            else:
                widget = self.make_numeric_widget(
                    table, i, form, name, part, param, order)

        elif param.type == fracttypes.Float or \
                param.type == fracttypes.Complex or \
                param.type == fracttypes.Hyper:

            widget = self.make_numeric_widget(
                table, i, form, name, part, param, order)
        elif param.type == fracttypes.Bool:
            widget = self.make_bool_widget(
                form, name, param, order)
        elif param.type == fracttypes.Color:
            widget = self.make_color_widget(
                table, i, form, name, param, order)
        elif param.type == fracttypes.Image:
            # skip image params for now
            return
        else:
            raise "Unsupported parameter type"

        table.attach(widget, 1, i, 1, 1)

    def add_complex_formula_setting(
            self, table, i, form, name, param, order, param_type):

        widget = self.make_numeric_entry(form, param, order)
        widget.set_hexpand(True)
        table.attach(widget, 1, i, 1, 1)

        widget = self.make_numeric_entry(form, param, order + 1)
        table.attach(widget, 1, i + 1, 1, 1)

        name = self.param_display_name(name, param)
        fway = fourway.T(name)
        tip = self.param_tip(name, param)
        fway.set_tooltip_text(tip)
        fway.set_hexpand(True)

        fway.connect('value-changed', self.fourway_released, order, form)

        if self.parent:
            fway.connect(
                'value-slightly-changed',
                self.parent.on_drag_param_fourway, order, param_type)

        table.attach(fway, 0, i, 1, 2)

    def fourway_released(self, widget, x, y, order, form):
        form.nudge_param(order, x, y)

    def construct_function_menu(self, param, form):
        funclist = sorted(form.formula.symbols.available_param_functions(
            param.ret, param.args))
        return funclist

    def set_nthreads(self, n):
        if self.nthreads != n:
            self.nthreads = n
            self.changed()

    def error(self, msg, err):
        print(self, self.parent)
        if self.parent:
            self.parent.show_error_message(msg, err)
        else:
            print("Error: %s : %s" % (msg, err))

    def warn(self, msg):
        if self.parent:
            self.parent.show_warning(msg)
        else:
            print("Warning: ", msg)

    def add_formula_function(self, table, i, name, param, form):
        label = Gtk.Label.new(self.param_display_name(name, param))
        label.set_halign(Gtk.Align.END)
        label.set_valign(Gtk.Align.CENTER)
        table.attach(label, 0, i, 1, 1)

        funclist = self.construct_function_menu(param, form)
        widget = utils.create_option_menu(funclist)

        formula = form.formula

        def set_selected_function():
            try:
                selected_func_name = form.get_func_value(name)
                index = funclist.index(selected_func_name)
            except ValueError as err:
                # func.cname not in list
                # print "bad cname"
                return

            utils.set_selected(widget, index)

        def set_fractal_function(om, f, param, formula):
            index = utils.get_selected(om)
            if index != -1:
                # this shouldn't be necessary but I got weird errors
                # trying to reuse the old funclist
                list = sorted(formula.symbols.available_param_functions(
                    param.ret, param.args))

                fname = list[index]
                f.set_func(param, fname, formula)

        set_selected_function()

        widget.update_function = set_selected_function

        widget.connect('changed', set_fractal_function, self, param, formula)

        table.attach(widget, 1, i, 1, 1)

    def create_maxiter_widget(self, table, i):
        label = Gtk.Label(label="_Max Iterations")
        label.set_halign(Gtk.Align.END)
        label.set_valign(Gtk.Align.CENTER)
        label.set_use_underline(True)
        table.attach(label, 0, i, 1, 1)

        widget = Gtk.Entry()
        widget.set_activates_default(True)

        def set_entry(*args):
            widget.set_text("%d" % self.f.maxiter)

        def set_fractal(*args):
            try:
                try:
                    i = int(widget.get_text())
                    self.set_maxiter(i)
                except ValueError as err:
                    msg = "Invalid value '%s': must be a number" % \
                          widget.get_text()
                    utils.idle_add(self.warn, msg)
            except Exception as exn:
                print(exn)
            return False

        set_entry(self)

        self.connect('parameters-changed', set_entry)
        self.connect('iters-changed', set_entry)
        widget.connect('focus-out-event', set_fractal)

        label.set_mnemonic_widget(widget)
        table.attach(widget, 1, i, 1, 1)
        return i + 1

    def populate_formula_settings(self, table, param_type, row=0):
        # create widget to fiddle with this fractal's settings
        form = self.f.get_form(param_type)
        formula = form.formula

        if param_type == 0:
            row = self.create_maxiter_widget(table, row)

        params = formula.symbols.parameters()
        op = formula.symbols.order_of_params()

        keys = sorted(params.keys())
        for name in keys:
            param = params[name]
            if isinstance(param, function.Func):
                self.add_formula_function(table, row, name, param, form)
            else:
                if param.type == fracttypes.Complex:
                    self.add_complex_formula_setting(
                        table, row, form, name, param,
                        op[name],
                        param_type)
                    row += 1
                elif param.type == fracttypes.Hyper:
                    suffixes = [" (re)", " (i)", " (j)", " (k)"]
                    for j in range(4):
                        self.add_formula_setting(
                            table, row + j, form, name, suffixes[j],
                            param, op[name] + j)
                    row += 3
                elif param.type == fracttypes.Color:
                    self.add_formula_setting(
                        table, row, form, name, "",
                        param, op[name])
                    row += 3
                elif param.type == fracttypes.Gradient:
                    # FIXME
                    pass
                else:
                    self.add_formula_setting(
                        table, row, form, name, "", param, op[name])
            row += 1
        return table

    def set_size(self, new_width, new_height):
        try:
            Hidden.set_size(self, new_width, new_height)
            self.widget.set_size_request(new_width, new_height)
        except MemoryError as err:
            utils.idle_add(self.warn, str(err))

    def draw_image(self, aa=None, auto_deepen=None):
        try:
            Hidden.draw_image(self, aa, auto_deepen)
        except fracttypes.TranslationError as err:
            advice = _(
                "\nCheck that your compiler settings and formula file are correct.")
            utils.idle_add(self.error,
                           _("Error compiling fractal:"),
                           err.msg + advice)
            return

    def onMotionNotify(self, widget, event):
        x, y = self.float_coords(event.x, event.y)
        self.emit('pointer-moved', self.button, x, y)

        if not self.notice_mouse:
            return

        self.newx, self.newy = event.x, event.y

        dy = int(abs(self.newx - self.x) * float(self.height) / self.width)
        if(self.newy < self.y or (self.newy == self.y and self.newx < self.x)):
            dy = -dy
        self.newy = self.y + dy

        # create a dummy Cairo context to calculate the affected bounding box
        surface = cairo.ImageSurface(cairo.FORMAT_A1, self.width, self.height)
        cairo_ctx = cairo.Context(surface)
        cairo_ctx.set_line_width(T.SELECTION_LINE_WIDTH)
        if self.selection_rect:
            cairo_ctx.rectangle(*self.selection_rect)

        self.selection_rect = [
            int(min(self.x, self.newx)),
            int(min(self.y, self.newy)),
            int(abs(self.newx - self.x)),
            int(abs(self.newy - self.y))]

        cairo_ctx.rectangle(*self.selection_rect)
        x1, y1, x2, y2 = cairo_ctx.stroke_extents()

        self.widget.queue_draw_area(x1, y1, x2 - x1, y2 - y1)

    def onButtonPress(self, widget, event):
        self.x = event.x
        self.y = event.y
        self.newx = self.x
        self.newy = self.y
        self.button = event.button
        if self.button == 1:
            self.notice_mouse = True

    def set_paint_mode(self, isEnabled, colorsel):
        self.paint_mode = isEnabled
        self.paint_color_sel = colorsel

    def get_paint_color(self):
        color = self.paint_color_sel.get_current_color()
        return (color.red / 65535.0, color.green /
                65535.0, color.blue / 65535.0)

    def onPaint(self, x, y):
        # obtain index
        fate = self.image.get_fate(int(x), int(y))
        if not fate:
            return

        index = self.image.get_color_index(int(x), int(y))

        # obtain a color
        (r, g, b) = self.get_paint_color()

        # update colormap
        grad = self.f.get_gradient()

        (is_solid, color_type) = fate
        if color_type == 0x20:
            print("update fate")
            color_type = 1  # FATE_UNKNOWN is treated as INSIDE
        if is_solid:
            self.f.solids[color_type] = (
                int(r * 255.0), int(g * 255.0), int(b * 255.0), 255)
        else:
            i = grad.get_index_at(index)
            if index > grad.segments[i].mid:
                alpha = grad.segments[i].right_color[3]
                grad.segments[i].right_color = [r, g, b, alpha]
            else:
                alpha = grad.segments[i].left_color[3]
                grad.segments[i].left_color = [r, g, b, alpha]

        self.changed(False)

    def filterPaintModeRelease(self, event):
        if self.paint_mode:
            if event.button == 1:
                if self.x == self.newx or self.y == self.newy:
                    self.onPaint(self.newx, self.newy)
                    return True

        return False

    def onButtonRelease(self, widget, event):
        self.widget.queue_draw()
        self.button = 0
        self.notice_mouse = False
        self.selection_rect.clear()
        if self.filterPaintModeRelease(event):
            return

        self.freeze()
        if event.button == 1:
            if self.x == self.newx or self.y == self.newy:
                zoom = 0.5
                x = self.x
                y = self.y
            else:
                zoom = (1 + abs(self.x - self.newx)) / float(self.width)
                x = 0.5 + (self.x + self.newx) / 2.0
                y = 0.5 + (self.y + self.newy) / 2.0

            # with shift held, don't zoom
            if hasattr(event, "state") and event.get_state() & Gdk.ModifierType.SHIFT_MASK:
                zoom = 1.0
            self.recenter(x, y, zoom)

        elif event.button == 2:
            (x, y) = (event.x, event.y)
            zoom = 1.0
            self.recenter(x, y, zoom)
            if self.is4D():
                self.flip_to_julia()

        else:
            if hasattr(event, "state") and event.get_state(
            ) & Gdk.ModifierType.CONTROL_MASK:
                zoom = 20.0
            else:
                zoom = 2.0
            (x, y) = (event.x, event.y)
            self.recenter(x, y, zoom)

        if self.thaw():
            self.changed()

    def redraw_rect(self, widget, cairo_ctx):
        result, r = Gdk.cairo_get_clip_rectangle(cairo_ctx)
        if result:
            x, y, w, h = r.x, r.y, r.width, r.height
        else:
            print("Skipping drawing because entire context clipped")
            return

        try:
            buf = self.image.image_buffer(x, y)
        except MemoryError as err:
            # suppress these errors
            return

        pixbuf = GdkPixbuf.Pixbuf.new_from_bytes(
            GLib.Bytes(buf),
            GdkPixbuf.Colorspace.RGB,
            False,
            8,
            min(self.width - x, w),
            min(self.height - y, h),
            self.width * 3)
        Gdk.cairo_set_source_pixbuf(cairo_ctx, pixbuf.copy(), x, y)
        cairo_ctx.paint()

        if self.selection_rect:
            cairo_ctx.set_source_rgb(1.0, 1.0, 1.0)
            cairo_ctx.set_line_width(T.SELECTION_LINE_WIDTH)
            cairo_ctx.rectangle(*self.selection_rect)
            cairo_ctx.stroke()


class Preview(T):
    def __init__(self, comp, width=120, height=90):
        T.__init__(self, comp, None, width, height)

    def onButtonRelease(self, widget, event):
        pass

    def error(self, msg, exn):
        # suppress errors from previews
        pass

    def stats_changed(self, s):
        pass


class SubFract(T):
    def __init__(self, comp, width=640, height=480):
        T.__init__(self, comp, None, width, height)
        self.master = None

    def set_master(self, master):
        self.master = master

    def onButtonRelease(self, widget, event):
        self.button = 0
        if self.master:
            self.master.set_fractal(self.copy_f())

    def onMotionNotify(self, widget, event):
        pass

    def error(self, msg, exn):
        # suppress errors from subfracts, if they ever happened
        # it would be too confusing
        pass



#!/usr/bin/env python3

# Trivial symbol table implementation

import copy
from collections import UserDict, UserList
import re
import math

from .function import Func
from .fracttypes import *
from . import ir


class OverloadList(UserList):
    def __init__(self, list, **kwds):
        UserList.__init__(self, list)
        self.pos = -1
        self._is_operator = kwds.get("operator")
        self.__doc__ = kwds.get("doc")
        self.declared = False

    def __copy__(self):
        copied_data = [copy.copy(x) for x in self.data]
        c = OverloadList(copied_data)
        c.pos = self.pos
        c._is_operator = self._is_operator
        c.__doc__ = self.__doc__
        return c

    def first(self):
        return self[0]

    def is_operator(self):
        return self._is_operator


def efl(fname, template, tlist, **kwds):
    'short-hand for expandFuncList - just reduces the amount of finger-typing'
    list = []
    for t in tlist:
        f = "Func(%s,\"%s\")" % (re.sub("_", str(t), template), fname)
        list.append(eval(f))
    return OverloadList(list, **kwds)


def cfl(template, tlist):
    list = []
    for t in tlist:
        f = re.sub("_", str(t), template)
        list.append(eval(f))
    return list


def mkf(args, ret, fname):
    # create a function
    return Func(args, ret, fname)


def mkfl(dict, name, lst, **kwds):
    "make a list of functions"
    fname = kwds.get("fname", name)  # fname overrides name if present
    # avoid having to provide list of one element
    if not isinstance(lst[0][0], list):
        lst = [lst]
    funclist = [mkf(x[0], x[1], fname) for x in lst]

    implicit = kwds.get("implicit", None)
    if implicit:
        for f in funclist:
            f.set_implicit_arg(implicit)

    dict[name] = OverloadList(funclist, **kwds)


class Alias:
    def __init__(self, realName):
        self.realName = realName
        self.pos = -1
        self.cname = None


def createDefaultDict():
    d = {
        "^": OverloadList([Func([Float, Float], Float, "pow"),
                           Func([Complex, Float], Complex, "pow"),
                           Func([Complex, Complex], Complex, "pow")],
                          operator=True,
                          doc='''Exponentiation operator. Computes x to the power y.'''),

        "t__neg": efl("neg", "[_], _", [Int, Float, Complex, Hyper]),

        # logical ops
        "&&": OverloadList(
            [Func([Bool, Bool], Bool, None)],
            doc="Logical AND.", operator=True),
        "||": OverloadList(
            [Func([Bool, Bool], Bool, None)],
            doc="Logical OR.", operator=True),
        "t__not": OverloadList(
            [Func([Bool], Bool, "not")],
            doc="Logical NOT.", operator=True),

        # predefined magic variables
        "t__h_pi": Alias("pi"),
        "t__h_rand": Alias("rand"),
        "t__h_random": Alias("rand"),
        "t__h_magn": Var(Float, doc="The magnification factor of the image. This is the number of times the image size has doubled, or ln(4.0/size)"),
        "t__h_center": Var(Complex, doc="Where the center of the image is located on the complex plane"),
        "rand": OverloadList(
            [Func([], Complex, "rand")],
            doc="Each time this is accessed, it returns a new pseudo-random complex number. This is primarily for backwards compatibility with Fractint formulas - use the random() function in new formulas."),

        "t__h_pixel": Alias("pixel"),
        "t__h_xypixel": Alias("pixel"),
        "pixel": Var(Complex, doc="The (X,Y) coordinates of the current point. When viewing the Mandelbrot set, this has a different value for each pixel. When viewing the Julia set, it remains constant for each pixel."),
        "t__h_z": Alias("z"),
        "z": Var(Complex),
        "t__h_index": Var(Float, doc="The point in the gradient to use for the color of this point."),
        "t__h_numiter": Var(Int, doc="The number of iterations performed."),
        "t__h_maxiter": Alias("maxiter"),
        "t__h_maxit": Alias("maxiter"),
        "maxit": Alias("maxiter"),
        "maxiter": Var(Int, "The maximum number of iterations set by the user."),
        "pi": Var(Float, math.pi, doc="The constant pi, 3.14159..."),
        "t__h_tolerance": Var(Float, doc="10% of the distance between adjacent pixels."),
        "t__h_zwpixel": Var(Complex, doc="The (Z,W) coordinates of the current point. (See #pixel for the other two coordinates.) When viewing the Mandelbrot set, this remains constant for each pixel on the screen; when viewing the Julia set, it's different for each pixel. Initialize z to some function of this to take advantage of 4D drawing."),
        "t__h_solid": Var(Bool, doc="Set this to true in a coloring function to use the solid color rather than the color map."),
        "t__h_color": Var(Color, doc="Set this from a coloring function to directly set the color instead of using a gradient"),
        "t__h_fate": Var(Int, doc="The fate of a point can be used to distinguish between different basins of attraction or whatever you like. Set this to a number from 2 to 128 to indicate that a different 'fate' has befallen this point. 0 indicates the point has diverged, 1 that it has been trapped, >1 whatever you like. Can only be usefully updated in the #final section."),
        "t__h_inside": Var(Bool, doc="Set this in the final section of a formula to override whether a point is colored with the inside or outside coloring algorithm. This is mainly useful in conjuction with #fate.")
    }

    # extra shorthand to make things as short as possible
    def f(name, list, **kwds):
        mkfl(d, name, list, **kwds)

    # standard functions
    f("bool",
      [[Bool], Bool],
      doc="""Construct a boolean. It's not really required (bool x = bool(true) is just the same as bool x = true) but is included for consistency.""")

    f("int",
      [[Int], Int],
      doc="""Construct an integer. To convert a float to an int, use floor, ceil, round or trunc instead.""")

    f("float",
      [[Float], Float],
      doc="""Construct a floating-point number.""")

    f("color",
      [[Float, Float, Float, Float], Color],
      doc="""Constructs a new color from floating point red, green, blue and alpha
     components. Equivalent to rgba.""")

    f("complex",
      [[Float, Float], Complex],
      doc='''Construct a complex number from two real parts.
      complex(a,b) is equivalent to (a,b).''')

    f("hyper",
      [[[Float, Float, Float, Float], Hyper], [[Complex, Complex], Hyper]],
      doc='''Construct a hypercomplex number with a real and 3 imaginary parts.
      Can be passed either 2 complex numbers or 4 floating-point numbers.
      hyper(a,b,c,d) is equivalent to the shorthand (a,b,c,d).''')

    f("sqr",
      cfl("[_] , _", [Int, Float, Complex, Hyper]),
      doc="Square the argument. sqr(x) is equivalent to x*x or x^2.")

    # f("cube",
    #  cfl("[_] , _", [Int, Float, Complex]),
    #  doc="Cube the argument. cube(x) is equivalent to x*x*x or x^3.")

    f("ident",
      cfl("[_] , _", [Int, Float, Complex, Bool, Hyper]),
      doc='''Do nothing. ident(x) is equivalent to x.
      This function is useless in normal formulas but
      comes in useful as a value for a function parameter
      to a formula. For example, a general formula like z = @fn1(z*z)+c
      can be set back to a plain Mandelbrot by setting fn1 to ident.
      Note: ident() is compiled out so there\'s no speed penalty involved.''')

    f("conj",
      cfl("[_] , _", [Complex, Hyper]),
      doc="The complex conjugate. conj(a,b) is equivalent to (a,-b).")

    f("flip",
      cfl("[_] , _", [Complex, Hyper]),
      doc='''Swap the real and imaginary parts of a complex number.
      flip(a,b) = (b,a).''')

    f("real",
      [[[Complex], Float], [[Hyper], Float]],
      doc='''Extract the real part of a complex or hypercomplex number.
      real(a,b) = a.
      real() is unusual in that it can be assigned to: real(z) = 7 changes
      the real part of z.''')

    f("real2",
      [[Complex], Float],
      doc='''The square of the real part of a complex number.
      real2(a,b) = a*a.
      While not a generally useful function, this is provided to ease porting
      of files from older Gnofract 4D versions.''')

    f("imag",
      [[[Complex], Float], [[Hyper], Float]],
      doc='''Extract the imaginary part of a complex or hypercomplex number.
      imag(a,b) = b.
      imag() is unusual in that it can be assigned to: imag(z) = 7 changes
      the imag part of z.''')

    f("imag2",
      [[Complex], Float],
      doc='''The square of the imaginary part of a complex number.
      real2(a,b) = b*b.
      While not a generally useful function, this is provided to ease porting
      of files from older Gnofract 4D versions.''')

    f("hyper_ri",
      [[Hyper], Complex],
      doc='''The real and imaginary parts of a hypercomplex number.
      Can be assigned to. hyper_ri(a,b,c,d) = (a,b).''')

    f("hyper_jk",
      [[Hyper], Complex],
      doc='''The 3rd and 4th parts of a hypercomplex number.
      Can be assigned to. hyper_jk(a,b,c,d) = (c,d).''')

    f("hyper_j",
      [[Hyper], Float],
      doc='''The 3rd component of a hypercomplex number. Can be assigned to.
      hyper_j(a,b,c,d) = c.''')

    f("hyper_k",
      [[Hyper], Float],
      doc='''The 4th component of a hypercomplex number. Can be assigned to.
      hyper_k(a,b,c,d) = d.''')

    f("red",
      [[Color], Float],
      doc='''The red component of a color. Can be assigned to.''')

    f("green",
      [[Color], Float],
      doc='''The green component of a color. Can be assigned to.''')

    f("blue",
      [[Color], Float],
      doc='''The blue component of a color. Can be assigned to.''')

    f("alpha",
      [[Color], Float],
      doc='''The alpha component of a color. Can be assigned to.''')

    f("hue",
      [[Color], Float],
      doc='''The hue of a color.''')

    f("sat",
      [[Color], Float],
      doc='''The saturation of a color.''')

    f("lum",
      [[Color], Float],
      doc='''The luminance (or brightness) of a color.''')

    f("gradient",
      [[Float], Color],
      doc='''Look up a color from the default gradient.''')

    f("recip",
      cfl("[_] , _", [Float, Complex, Hyper]),
      doc='''The reciprocal of a number. recip(x) is equivalent to 1/x.
      Note that not all hypercomplex numbers have a proper reciprocal.''')

    f("trunc",
      [[[Float], Int], [[Complex], Complex]],
      doc='''Round towards zero.''')

    f("round",
      [[[Float], Int], [[Complex], Complex]],
      doc='''Round to the nearest number (0.5 rounds up).''')

    f("floor",
      [[[Float], Int], [[Complex], Complex]],
      doc='''Round down to the next lowest number.''')

    f("ceil",
      [[[Float], Int], [[Complex], Complex]],
      doc='''Round up to the next highest number.''')

    f("zero",
      cfl("[_], _ ", [Int, Float, Complex]),
      doc='''Returns zero.''')

    f("abs",
      cfl("[_], _", [Int, Float, Complex]),
      doc='''The absolute value of a number. abs(3) = abs(-3) = 3.
      abs() of a complex number is a complex number consisting of
      the absolute values of the real and imaginary parts, i.e.
      abs(a,b) = (abs(a),abs(b)).''')

    f("cabs",
      [[Complex], Float],
      doc='''The complex modulus of a complex number z.
      cabs(a,b) is equivalent to sqrt(a*a+b*b).
      This is also the same as sqrt(|z|)''')

    f("cmag",
      [[[Complex], Float], [[Hyper], Float]],
      doc='''The squared modulus of a complex or hypercomplex number z.
      cmag(a,b) is equivalent to a*a+b*b. This is the same as |z|.''')

    f("log",
      cfl("[_], _", [Float, Complex, Hyper]),
      doc='The natural log.')

    f("sqrt",
      cfl("[_], _", [Float, Complex, Hyper]),
      doc='''The square root.
      The square root of a negative float number is NaN
      (ie it is NOT converted to complex). Thus sqrt((-3,0)) != sqrt(-3).''')

    f("exp",
      cfl("[_], _", [Float, Complex, Hyper]),
      doc='exp(x) is equivalent to e^x')

    f("manhattan",
      [[Complex], Float],
      doc='''The Manhattan distance between the origin and complex number z.
      manhattan(a,b) is equivalent to abs(a) + abs(b).''')

    f("manhattanish",
      [[Complex], Float],
      doc='''A variant on Manhattan distance provided for backwards
      compatibility. manhattanish(a,b) is equivalent to a+b.''')

    f("manhattanish2",
      [[Complex], Float],
      doc='''A variant on Manhattan distance provided for backwards
      compatibility. manhattanish2(a,b) is equivalent to (a*a + b*b)^2.''')

    f("min",
      [[Float, Float], Float],
      doc='''Returns the smaller of its two arguments.''')

    f("max",
      [[Float, Float], Float],
      doc='''Returns the larger of its two arguments.''')

    f("max2",
      [[Complex], Float],
      doc='''max2(a,b) returns the larger of a*a or b*b. Provided for
      backwards compatibility.''')

    f("min2",
      [[Complex], Float],
      doc='''min2(a,b) returns the smaller of a*a or b*b. Provided for
      backwards compatibility.''')

    f("sin",
      cfl("[_], _", [Float, Complex, Hyper]),
      doc='trigonometric sine function.')

    f("cos",
      cfl("[_], _", [Float, Complex, Hyper]),
      doc='trigonometric sine function.')

    f("cosxx",
      cfl("[_], _", [Complex, Hyper]),
      doc='''Incorrect version of cosine function. Provided for backwards
      compatibility with equivalent wrong function in Fractint.''')

    f("tan",
      cfl("[_], _", [Float, Complex, Hyper]),
      doc='trigonometric sine function.')

    f("cotan",
      cfl("[_], _", [Float, Complex, Hyper]),
      doc="Trigonometric cotangent function.")

    f("sinh",
      cfl("[_], _", [Float, Complex, Hyper]),
      doc='Hyperbolic sine function.')

    f("cosh",
      cfl("[_], _", [Float, Complex, Hyper]),
      doc='Hyperbolic cosine function.')

    f("tanh",
      cfl("[_], _", [Float, Complex, Hyper]),
      doc='Hyperbolic tangent function.')

    f("cotanh",
      cfl("[_], _", [Float, Complex, Hyper]),
      doc='Hyperbolic cotangent function.')

    f("asin",
      cfl("[_], _", [Float, Complex, Hyper]),
      doc='Inverse sine function.')

    f("acos",
      cfl("[_], _", [Float, Complex, Hyper]),
      doc='Inverse cosine function.')

    f("atan",
      cfl("[_], _", [Float, Complex, Hyper]),
      doc='Inverse tangent function.')

    f("atan2",
      [[Complex], Float],
      doc='''The angle between this complex number and the real line,
      aka the complex argument.''')

    f("asinh",
      cfl("[_], _", [Float, Complex, Hyper]),
      doc='Inverse hyperbolic sine function.')

    f("acosh",
      cfl("[_], _", [Float, Complex, Hyper]),
      doc='Inverse hyperbolic cosine function.')

    f("atanh",
      cfl("[_], _", [Float, Complex, Hyper]),
      doc='Inverse hyperbolic tangent function.')

    # color functions
    f("blend",
      [[Color, Color, Float], Color],
      doc='Blend two colors together in the ratio given by the 3rd parameter.')

    f("compose",
      [[Color, Color, Float], Color],
      doc='''Composite the second color on top of the first, with opacity given
by the 3rd parameter.''')

    f("mergenormal",
      [[Color, Color], Color],
      doc='''Returns second color, ignoring first.''')

    f("mergemultiply",
      [[Color, Color], Color],
      doc='''Multiplies colors together. Result is always darker than either input.''')

    f("rgb",
      [[Float, Float, Float], Color],
      doc='''Create a color from three color components. The alpha channel is set to to 1.0 (=100%).''')

    f("rgba",
      [[Float, Float, Float, Float], Color],
      doc='Create a color from three color components and an alpha channel.')

    f("hsl",
      [[Float, Float, Float], Color],
      doc='''Create a color from hue, saturation and lightness components. The alpha channel is set to to 1.0 (=100%).''')

    f("hsla",
      [[Float, Float, Float, Float], Color],
      doc='''Create a color from hue, saturation and lightness components and an alpha channel.''')

    f("hsv",
      [[Float, Float, Float], Color],
      doc='''Create a color from hue, saturation and value components. HSV is a similar color model to HSL but has a different valid range for brightness.''')

    f("_image",
      [[Image, Complex], Color],
      doc='''Look up a color from a 2D array of colors.''')

    f("_alloc",
      [[[VoidArray, Int, Int], VoidArray],
       [[VoidArray, Int, Int, Int], VoidArray],
       [[VoidArray, Int, Int, Int, Int], VoidArray],
       [[VoidArray, Int, Int, Int, Int, Int], VoidArray]],
      doc='''Allocate an array. First argument is element size in bytes, subsequent args are array sizes''')

    f("_write_lookup",
      # args are array, indexes, value. returns false if index is out of bounds
      [[[IntArray, Int, Int], Bool],
       [[IntArray, Int, Int, Int], Bool],
       [[IntArray, Int, Int, Int, Int], Bool],
       [[IntArray, Int, Int, Int, Int, Int], Bool],

       [[FloatArray, Int, Float], Bool],
       [[FloatArray, Int, Int, Float], Bool],
       [[FloatArray, Int, Int, Int, Float], Bool],
       [[FloatArray, Int, Int, Int, Int, Float], Bool],

       [[ComplexArray, Int, Complex], Bool],
       [[ComplexArray, Int, Int, Complex], Bool],
       [[ComplexArray, Int, Int, Int, Complex], Bool],
       [[ComplexArray, Int, Int, Int, Int, Complex], Bool],

       ],
      doc='''Write a value into an array''')

    f("_read_lookup",
      # args are array, indexes, value
      [[[IntArray, Int], Int],
       [[IntArray, Int, Int], Int],
       [[IntArray, Int, Int, Int], Int],
       [[IntArray, Int, Int, Int, Int], Int],

       [[FloatArray, Int], Float],
       [[FloatArray, Int, Int], Float],
       [[FloatArray, Int, Int, Int], Float],
       [[FloatArray, Int, Int, Int, Int], Float],

       [[ComplexArray, Int], Complex],
       [[ComplexArray, Int, Int], Complex],
       [[ComplexArray, Int, Int, Int], Complex],
       [[ComplexArray, Int, Int, Int, Int], Complex],

       ],
      doc='''Read a value out of an array''')

    # operators

    f("+",
      cfl("[_,_] , _", [Int, Float, Complex, Hyper, Color]),
      fname="add",
      operator=True,
      doc='Adds two numbers together.')

    f("-",
      cfl("[_,_] , _", [Int, Float, Complex, Hyper, Color]),
      fname="sub",
      operator=True,
      doc='Subtracts two numbers')

    f("*",
      cfl("[_,_] , _", [Int, Float, Complex, Hyper]) +
      cfl("[_, Float], _", [Hyper, Color]),
      fname="mul",
      operator=True,
      doc='''Multiplication operator.''')

    f("/",
      [
          [[Float, Float], Float],
          [[Complex, Float], Complex],
          [[Complex, Complex], Complex],
          [[Hyper, Float], Hyper],
          [[Color, Float], Color]
      ],
      fname="div",
      operator=True,
      doc='''Division operator''')

    f("!=",
      cfl("[_,_] , Bool", [Int, Float, Complex, Bool]),
      fname="noteq",
      operator=True,
      precedence=3,
      doc='''Inequality operator. Compare two values and return true if
      they are different.''')

    f("==",
      cfl("[_,_] , Bool", [Int, Float, Complex, Bool]),
      fname="eq",
      operator=True,
      precedence=3,
      doc='''Equality operator. Compare two values and return true if they are
      the same.''')

    # fixme - issue a warning for complex compares
    f(">",
      cfl("[_,_], Bool", [Int, Float, Complex]),
      fname="gt",
      operator=True,
      precedence=3,
      doc='''Greater-than operator. Compare two values and return true if the first is greater than the second.''')

    f(">=",
      cfl("[_,_], Bool", [Int, Float, Complex]),
      fname="gte",
      operator=True,
      precedence=3,
      doc='''Greater-than-or-equal operator. Compare two values and return true if the first is greater than or equal to the second.''')

    f("<",
      cfl("[_,_], Bool", [Int, Float, Complex]),
      fname="lt",
      operator=True,
      precedence=3,
      doc='''Less-than operator. Compare two values and return true if the first is less than the second.''')

    f("<=",
      cfl("[_,_], Bool", [Int, Float, Complex]),
      fname="lte",
      operator=True,
      precedence=3,
      doc='''Less-than-or-equal operator. Compare two values and return true if the first is less than or equal to the second.''')

    f("%",
      cfl("[_,_] , _", [Int, Float]),
      fname="mod",
      operator=True,
      doc='''Modulus operator. Computes the remainder when x is divided by y. Not to be confused with the complex modulus.'''),

    # predefined parameters
    for p in range(1, 7):
        name = "p%d" % p
        d[name] = Alias("t__a_" + name)
        d["t__a_" + name] = Var(Complex,
                                doc="Predefined parameter used by Fractint formulas")

    # predefined functions
    for p in range(1, 5):
        name = "fn%d" % p
        d[name] = Alias("t__a_" + name)
        d["t__a_" + name] = OverloadList(
            [Func([Complex], Complex, "ident")],
            doc="Predefined function parameter used by Fractint formulas")

    # predefined gradient-related vars and functions
    tfunc = Func([Float], Float, "ident")
    tfunc.caption = ir.Const("Transfer Function", None, String)
    transfer = OverloadList([tfunc])
    d["t__a__transfer"] = transfer

    d["t__a__gradient"] = Var(Gradient)

    for (k, v) in list(d.items()):
        if hasattr(v, "cname") and v.cname is None:
            v.cname = k

    return d


def mangle(k, prefix=""):
    l = k.lower()
    if l[0] == '#':
        l = "t__h_" + prefix + l[1:]
    elif l[0] == '@':
        l = "t__a_" + prefix + l[1:]
    return l


class T(UserDict):
    default_dict = createDefaultDict()

    def __init__(self, prefix=""):
        UserDict.__init__(self)
        self.reset()
        self.nextlabel = 0
        self.nextTemp = 0
        self.nextEnum = 0
        self.nextParamSlot = 0
        self.var_params = []
        self.prefix = prefix
        self.temp_prefix = "t__%s" % prefix

    def __copy__(self):
        c = T(self.prefix)
        c.nextlabel = self.nextlabel
        c.nextTemp = self.nextTemp
        for k in list(self.data.keys()):
            c.data[k] = copy.copy(self.data[k])

        return c

    def merge(self, other):
        # self = union(self,other)
        # any clashes are won by self
        for k in list(other.data.keys()):
            # print "key",k
            if self.data.get(k) is None:
                # print "don't already have", k
                if self.is_param(k):
                    # print "update param", k
                    new_key = self.insert_prefix(other.prefix, k)
                    new_val = copy.copy(other.data[k])
                    self.data[new_key] = new_val
                    if isinstance(new_val, Var):
                        new_val.param_slot += self.nextParamSlot
                else:
                    self.data[k] = copy.copy(other.data[k])
            elif hasattr(self.data[k], "cname") and \
                    hasattr(other.data[k], "cname") and \
                    self.data[k].cname != other.data[k].cname:
                # print "have", k
                new_key = self.insert_prefix(other.prefix, k)
                new_val = copy.copy(other.data[k])
                self.data[new_key] = new_val
                if self.is_param(k) and isinstance(new_val, Var):
                    new_val.param_slot += self.nextParamSlot

        self.nextParamSlot += other.nextParamSlot

    def has_user_key(self, key):
        return mangle(key) in self.data

    def has_key(self, key):
        if mangle(key) in self.data:
            return True
        return mangle(key) in self.default_dict

    def is_user(self, key):
        val = self.data.get(mangle(key), None)
        if val is None:
            val = self.default_dict.get(mangle(key))
        return val.pos != -1

    def insert_prefix(self, prefix, key):
        if key[0:5] == "t__a_":
            return "t__a_" + prefix + key[5:]
        if key[0:3] == "t__":
            return "t__" + prefix + key[3:]
        return prefix + key

    def is_param(self, key):
        return key[0:5] == 't__a_'

    def is_private(self, key):
        return key[0:3] == "t__"

    def mangled_name(self, key):
        k = mangle(key)
        return k

    def realName(self, key):
        ' returns mangled key even if var not present for test purposes'
        k = mangle(key)
        return self._realName(k)

    def _realName(self, k):
        val = self.data.get(k, None)
        if val is None:
            val = self.default_dict.get(k)
        if isinstance(val, Alias):
            val = self.default_dict.get(val.realName)
        if val is not None:
            if val.cname is None:
                # print k
                raise Exception("argh" + k)
            return val.cname
        return k

    def get(self, key):
        try:
            return self[key]
        except KeyError:
            return None

    def __contains__(self, key):
        k = mangle(key)
        val = self.data.get(k, None)
        if val is None:
            val = self.default_dict.get(k, None)

        return val

    def __getitem__(self, key):
        k = mangle(key)
        val = self.data.get(k, None)
        if val is None:
            val = self.default_dict[k]
            if isinstance(val, Alias):
                key = val.realName
                return self.__getitem__(key)

            val = copy.copy(val)
            self.data[k] = val
            if self.is_param(k) and isinstance(val, Var) and \
                    val.param_slot == -1:
                self.record_param(val)
        return val

    def is_builtin(self, key):
        return key[0] == '#'

    def clashes_with_private(self, mangled_key, old_key):
        if mangled_key.find("t__", 0, 3) != 0:
            return False
        if old_key[0] == '@':
            return False
        return True

    def record_param(self, value):
        if value.cname == "z":
            print(value)
            assert False
        value.param_slot = self.nextParamSlot
        self.nextParamSlot += slotsForType(value.type)
        self.var_params.append(value)

    def __setitem__(self, key, value):
        k = mangle(key)
        if k in self.data:
            pre_type = self.data[k].type
            if pre_type != value.type:
                l = self.data[k].pos
                msg = ("was already defined as %s on line %d" %
                       (strOfType(pre_type), l))
                raise KeyError("symbol '%s' %s" % (key, msg))
            return
        elif k in T.default_dict:
            pre_var = T.default_dict[k]
            #print("in default: %s" % k)
            if isinstance(pre_var, OverloadList):
                msg = "is predefined as a function"
                raise KeyError("symbol '%s' %s" % (key, msg))
            else:
                if pre_var.type != value.type:
                    msg = "is predefined as %s" % \
                          strOfType(T.default_dict[k].type)
                    raise KeyError("symbol '%s' %s" % (key, msg))

                if self.is_param(k):
                    self.record_param(pre_var)
            return
        elif self.is_builtin(key):
            msg = "symbol '%s': only predefined symbols can begin with #" % key
            raise KeyError(msg)
        elif self.clashes_with_private(k, key):
            raise KeyError(
                "symbol '%s': no symbol starting with t__ is allowed" % key)
        self.data[k] = value
        if self.is_param(k) and isinstance(value, Var):
            # print "recording",k
            self.record_param(value)
        else:
            pass  # print "not recording",k

        if hasattr(value, "cname") and value.cname is None:
            value.cname = self.insert_prefix(self.prefix, k)

    def ensure(self, name, var):
        # make sure an item is referred to in main dict
        self.__setitem__(name, var)
        self.__getitem__(name)

    def parameters(self, varOnly=False):
        params = {}
        for (name, sym) in list(self.data.items()):
            if self.is_param(name):
                if not varOnly or isinstance(sym, Var):
                    try:
                        params[name] = sym.first()
                    except AttributeError:
                        print(sym, name)
                        raise

        return params

    def demangle(self, name):
        # remove most obvious mangling.
        # because of case-folding, demangle(mangle(s)) != s
        if name[:3] == "t__":
            name = name[3:]

        if name[:2] == "a_":
            name = "@" + name[2:]
        elif name[:2] == "h_":
            name = "#" + name[2:]

        return name

    def is_direct(self):
        return self.has_user_key("#color")

    def param_names(self):
        params = self.parameters()

        names = []
        for (name, param) in list(params.items()):
            if isinstance(param, Var):
                names.append(self.demangle(name))

        return names

    def func_names(self):
        params = self.parameters()

        func_names = []
        for (name, param) in list(params.items()):
            if isinstance(param, Func):
                func_names.append(self.demangle(name))
        return func_names

    def available_param_functions(self, ret, args):
        # a list of all function names which take args of type 'args'
        # and return 'ret' (for GUI to select a function)
        flist = []
        for (name, func) in list(self.default_dict.items()):
            try:
                for f in func:
                    if f.ret == ret and f.args == args and \
                        not self.is_private(name) and \
                            not func.is_operator():
                        flist.append(name)
            except TypeError:
                # wasn't a list
                pass

        return flist

    def order_of_params(self):
        # a hash which maps param name -> order in input list
        p = self.parameters(True)

        op = {}
        for k in list(p.keys()):
            op[k] = self[k].param_slot

        op["__SIZE__"] = self.nextParamSlot

        return op

    def type_of_params(self):
        # an array from param order -> type
        p = self.parameters(True)

        tp = [None] * self.nextParamSlot
        for k in list(p.keys()):
            i = self[k].param_slot
            t = p[k].type
            if t == Complex:
                tp[i:i + 2] = [Float, Float]
            elif t == Hyper or t == Color:
                tp[i:i + 4] = [Float, Float, Float, Float]
            elif t == Float:
                tp[i] = Float
            elif t == Int:
                tp[i] = Int
            elif t == Bool:
                tp[i] = Bool
            elif t == Gradient:
                tp[i] = Gradient
            elif t == Image:
                tp[i] = Image
            else:
                raise ValueError("Unknown param type %s for %s" % (t, k))
        #assert not None in tp
        return tp

    def default_params(self):
        op = self.order_of_params()
        defaults = [0.0] * op["__SIZE__"]

        for (k, i) in list(op.items()):
            param = self.get(k)
            if not param:
                continue
            defval = getattr(param, "default", None)
            if param.type == Complex:
                if not defval:
                    dv = default_value(param.type)
                    defaults[i:i + 2] = dv
                else:
                    defaults[i] = defval.value[0].value
                    defaults[i + 1] = defval.value[1].value
            elif param.type == Hyper or param.type == Color:
                for j in range(len(defval.value)):
                    defaults[i + j] = defval.value[j].value
            elif param.type == Image:
                defaults[i] = 0
            else:
                if not defval:
                    defaults[i] = default_value(param.type)
                else:
                    defaults[i] = defval.value
        return defaults

    def set_std_func(self, func, fname):
        # repoint parameter @func to use fname next time we compile
        func.set_func(fname)

    def __delitem__(self, key):
        del self.data[mangle(key)]

    def reset(self):
        self.data = {}

    def newLabel(self):
        label = "%slabel%d" % (self.prefix, self.nextlabel)
        self.nextlabel += 1
        return label

    def newTemp(self, type):
        # "%s%d" % (self.temp_prefix, self.nextTemp)
        name = self.temp_prefix + str(self.nextTemp)
        self.nextTemp += 1

        # bypass normal setitem because that checks for t__
        self.data[name] = Temp(type, name)

        return name

    def newEnum(self, name, val, pos):
        var = Var(Int, val, pos)
        # set cname because the enum value may not be a valid C identifier
        var.cname = "enum%s%d" % (self.prefix, self.nextEnum)
        self.nextEnum += 1
        self["__enum_" + name] = var
        return var



# This file contains utility classes and functions used by the GUI
# Many of them 'cover up' differences between pygtk versions - these
# follow the general pattern
#
# try:
#    do new thing
# except:
#    fall back to the 'old way'

import os
import inspect

import gi
gi.require_version('Gdk', '3.0')
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, Gdk, Gio, GLib

from . import hig

threads_enabled = False
break_new_things = False


def threads_enter():
    if threads_enabled:
        Gdk.threads_enter()


def threads_leave():
    if threads_enabled:
        Gdk.threads_leave()


def idle_wrapper(callable, *args):
    threads_enter()
    callable(*args)
    threads_leave()


def idle_add(callable, *args):
    """A wrapper around GObject.idle_add which wraps the callback in
    threads_enter/threads_leave if required"""
    GLib.idle_add(idle_wrapper, callable, *args)


def input_add(fd, cb):
    return GLib.io_add_watch(fd, GLib.PRIORITY_DEFAULT,
                             GLib.IO_IN | GLib.IO_HUP | GLib.IO_PRI, cb)


def find_in_path(exe):
    # find an executable along PATH env var
    pathstring = os.environ["PATH"]
    if not pathstring:
        return None
    paths = pathstring.split(":")
    for path in paths:
        full_path = os.path.join(path, exe)
        if os.path.exists(full_path):
            return full_path
    return None


def stack_trace():
    stack = inspect.stack()
    str = ""
    for frame in stack[1:]:
        (frame_obj, filename, line, funcname, context, context_index) = frame
        try:
            args = inspect.formatargvalues(*inspect.getargvalues(frame_obj))
        except Exception:
            args = "<unavailable>"

        frame_desc = "%s(%s)\t\t%s(%s)\n" % (filename, line, funcname, args)
        str += frame_desc
    return str


def get_directory_chooser(title, parent):
    chooser = Gtk.FileChooserDialog(
        title=title,
        transient_for=parent,
        action=Gtk.FileChooserAction.SELECT_FOLDER)

    chooser.add_buttons(
        Gtk.STOCK_OK, Gtk.ResponseType.OK,
        Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL)

    return chooser


def get_file_chooser_extra_widget(chooser):
    return chooser.get_extra_widget()


def set_file_chooser_filename(chooser, name):
    if name:
        chooser.set_current_folder(os.path.abspath(os.path.dirname(name)))
        chooser.set_current_name(os.path.basename(name))


def create_option_menu(items):
    widget = Gtk.ComboBoxText()
    for item in items:
        widget.append_text(item)

    return widget


def set_menu_from_list(menu, items):
    model = Gtk.ListStore(str)
    for item in items:
        model.append((item,))
    menu.set_model(model)


def add_menu_item(menu, item):
    menu.append_text(item)


def set_selected(menu, i):
    menu.set_active(int(i))


def get_selected(menu):
    return menu.get_active()


def get_selected_value(menu):
    iter = menu.get_active_iter()
    if not iter:
        return None
    val = menu.get_model().get_value(iter, 0)
    return val


def set_selected_value(menu, val):
    model = menu.get_model()
    i = 0
    iter = model.get_iter_first()
    while iter is not None:
        item = model.get_value(iter, 0)
        if item == val:
            menu.set_active(i)
            return
        iter = model.iter_next(iter)
        i += 1


def create_color(r, g, b):
    return Gdk.RGBA(int(r * 65535), int(g * 65535), int(b * 65535))


def floatColorFrom256(rgba):
    return [rgba[0] / 255.0, rgba[1] / 255.0, rgba[2] / 255.0, rgba[3] / 255.0]


def color256FromFloat(r, g, b, color):
    return (int(r * 255), int(g * 255), int(b * 255), color[3])


def launch_browser(url, window):
    try:
        Gio.AppInfo.launch_default_for_uri(url)
    except GLib.Error as err:
        d = hig.ErrorAlert(
            primary=_("Error launching browser"),
            secondary=_(f"Try copying the URL '{url}' manually to a browser window.\n"
                        f"{err.message}"),
            transient_for=window)
        d.run()
        d.destroy()


class ColorButton(Gtk.ColorButton):
    def __init__(self, rgb, changed_cb, is_left):
        Gtk.ColorButton.__init__(self)
        self.set_color(rgb)
        self.changed_cb = changed_cb
        self.is_left = is_left

        self.connect('color-set', self.on_color_set)

    def on_color_set(self, widget):
        self.color_changed(self.get_color())

    def set_color(self, rgb):
        self.color = create_color(rgb[0], rgb[1], rgb[2])
        Gtk.ColorButton.set_rgba(self, self.color)

    def color_changed(self, color):
        self.color = color
        self.changed_cb(
            color.red / 65535.0,
            color.green / 65535.0,
            color.blue / 65535.0,
            self.is_left)



import struct

from . import fract4dc


def parse(type, buffer):
    if type == fract4dc.MESSAGE_TYPE_ITERS:
        return Iters(buffer)
    if type == fract4dc.MESSAGE_TYPE_IMAGE:
        return Image(buffer)
    if type == fract4dc.MESSAGE_TYPE_PROGRESS:
        return Progress(buffer)
    if type == fract4dc.MESSAGE_TYPE_STATUS:
        return Status(buffer)
    if type == fract4dc.MESSAGE_TYPE_TOLERANCE:
        return Tolerance(buffer)
    if type == fract4dc.MESSAGE_TYPE_STATS:
        return Stats(buffer)


class T:
    pass


class Iters(T):
    def __init__(self, buffer):
        (self.iterations,) = struct.unpack("i", buffer)

    def get_name(self):
        return "Iters"
    name = property(get_name)

    def show(self):
        return "Iters: %d" % self.iterations


class Image(T):
    def __init__(self, buffer):
        (self.x, self.y, self.w, self.h) = struct.unpack("4i", buffer)

    def get_name(self):
        return "Image"
    name = property(get_name)

    def show(self):
        return "Image: (%d,%d) (%d,%d)" % (self.x, self.y, self.w, self.h)


class Progress(T):
    def __init__(self, buffer):
        (p,) = struct.unpack("i", buffer)
        self.progress = float(p)

    def show(self):
        return "Progress: %s" % self.progress

    def get_name(self):
        return "Progress"
    name = property(get_name)


class Status(T):
    def __init__(self, buffer):
        (self.status,) = struct.unpack("i", buffer)

    def show(self):
        return "Status: %d" % self.status

    def get_name(self):
        return "Status"
    name = property(get_name)


class Tolerance(T):
    def __init__(self, buffer):
        (self.tolerance,) = struct.unpack("d", buffer)

    def get_name(self):
        return "Tolerance"
    name = property(get_name)


class Stats(T):
    def fromList(list):
        instance = Stats()
        instance.iterations = list[0]
        instance.pixels = list[1]
        instance.pixels_calculated = list[2]
        instance.pixels_skipped = list[3]
        instance.pixels_skipped_wrong = list[4]
        instance.pixels_skipped_right = list[5]
        instance.pixels_inside = list[6]
        instance.pixels_outside = list[7]
        instance.pixels_periodic = list[8]
        return instance
    fromList = staticmethod(fromList)

    def __init__(self, buffer=None):
        if buffer:
            (self.iterations,
             self.pixels,
             self.pixels_calculated,
             self.pixels_skipped,
             self.pixels_skipped_wrong,
             self.pixels_skipped_right,
             self.pixels_inside,
             self.pixels_outside,
             self.pixels_periodic,
             dummy,
             dummy,
             dummy,
             dummy) = struct.unpack("13L", buffer)

    def _get_name(self):
        return "Stats"
    name = property(_get_name)

    def _get_percent_skipped(self):
        if self.pixels == 0:
            return 0.0
        return 100.0 * float(self.pixels_skipped) / self.pixels
    percent_skipped = property(_get_percent_skipped)

    def _get_percent_calculated(self):
        return 100.0 - self.percent_skipped
    percent_calculated = property(_get_percent_calculated)

    def _get_percent_skipped_wrong(self):
        if self.pixels_skipped == 0:
            return 0.0
        return 100.0 * float(self.pixels_skipped_wrong) / self.pixels_skipped
    percent_skipped_wrong = property(_get_percent_skipped_wrong)

    def _get_percent_inside(self):
        if self.pixels_calculated == 0:
            return 0.0
        return 100.0 * float(self.pixels_inside) / self.pixels_calculated
    percent_inside = property(_get_percent_inside)

    def _get_percent_periodic(self):
        if self.pixels_inside == 0:
            return 0.0
        return 100.0 * float(self.pixels_periodic) / self.pixels_inside
    percent_periodic = property(_get_percent_periodic)

    def _get_percent_outside(self):
        return 100.0 - self.percent_inside
    percent_outside = property(_get_percent_outside)

    def show(self):
        return (
            "Calculation Statistics:\n" +
            "Total pixels:%d\n" % self.pixels +
            "Calculated pixels:%d(%2g%%)\n" %
            (self.pixels_calculated, self.percent_calculated) +
            "  Inside pixels:%d(%2g%%)\n" % (self.pixels_inside, self.percent_inside) +
            "    Perodic pixels:%d(%2g%%)\n" % (self.pixels_periodic, self.percent_periodic) +
            "  Outside pixels:%d(%2g%%)\n" % (self.pixels_outside, self.percent_outside) +
            "Guessed pixels:%d(%2g%%)\n" % (
                self.pixels_skipped, self.percent_skipped)
        )

        # /out/per:\t%d\t%d\t%d\n" % \
        #    (self.pixels_inside, self.pixels_outside, self.pixels_periodic) +
        # "calc/skip:\t%d\t%d(%2g%%)\n" % \
        #    (self.pixels_calculated, self.pixels_skipped, self.percent_skipped) +
        # "skip right/wrong:\t%d\t%d(%2g%%)\n" % \
        #    (self.pixels_skipped_right, self.pixels_skipped_wrong, self.percent_skipped_wrong)
        # )



# whimsical feature to zoom in search of interesting items

import random
import gi

gi.require_version('Gtk', '3.0')
from gi.repository import Gtk

from . import dialog


class AutozoomDialog(dialog.T):
    def __init__(self, main_window, f):
        #pylint: disable=no-member
        dialog.T.__init__(
            self,
            _("Autozoom"),
            main_window,
            (Gtk.STOCK_CLOSE, Gtk.ResponseType.CLOSE)
        )

        self.f = f

        table = Gtk.Grid()
        table.set_column_spacing(5)
        table.set_row_spacing(5)
        self.vbox.add(table)

        self.zoombutton = Gtk.ToggleButton(label=_("Start _Zooming"))
        self.zoombutton.set_tooltip_text(
            _("Zoom into interesting areas automatically"))
        self.zoombutton.set_use_underline(True)
        self.zoombutton.connect('toggled', self.onZoomToggle)
        f.connect('status-changed', self.onStatusChanged)
        table.attach(self.zoombutton, 0, 0, 2, 1)

        self.minsize = 1.0E-13  # FIXME, should calculate this better

        self.minsize_entry = Gtk.Entry()
        self.minsize_entry.set_tooltip_text(
            _("Stop zooming when size of fractal is this small"))
        minlabel = Gtk.Label(label=_("_Min Size"))
        table.attach(minlabel, 0, 1, 1, 1)
        minlabel.set_use_underline(True)
        minlabel.set_mnemonic_widget(self.minsize_entry)

        def set_entry(*args):
            self.minsize_entry.set_text("%g" % self.minsize)

        def change_entry(*args):
            m = float(self.minsize_entry.get_text())
            if m != 0.0 and m != self.minsize:
                self.minsize = m
                set_entry()
            return False

        self.connect('focus-out-event', change_entry)
        set_entry()

        table.attach(self.minsize_entry, 1, 1, 1, 1)

        self.vbox.show_all()

    def onResponse(self, widget, id):
        self.zoombutton.set_active(False)
        self.hide()

    def onZoomToggle(self, *args):
        if self.zoombutton.get_active():
            self.zoombutton.get_child().set_text_with_mnemonic("Stop _Zooming")
            self.select_quadrant_and_zoom()
        else:
            self.zoombutton.get_child().set_text_with_mnemonic("Start _Zooming")

    def select_quadrant_and_zoom(self, *args):
        (wby2, hby2) = (self.f.width / 2, self.f.height / 2)
        (w, h) = (self.f.width, self.f.height)
        regions = [(0, 0, wby2, hby2),  # topleft
                   (wby2, 0, w, hby2),  # topright
                   (0, hby2, wby2, h),   # botleft
                   (wby2, hby2, w, h)]   # botright

        counts = [self.f.count_colors(r) for r in regions]
        m = max(counts)
        i = counts.index(m)

        # some level of randomness
        j = random.randrange(0, 4)
        if float(counts[j]) / counts[i] > 0.75:
            i = j

        # print "counts: %s max %d i %d" % (counts,m,i)

        # centers of each quadrant
        coords = [(1, 1), (3, 1), (1, 3), (3, 3)]

        (x, y) = coords[i]
        self.f.recenter(x * self.f.width / 4, y * self.f.height / 4, 0.75)

    def onStatusChanged(self, f, status_val):
        if status_val == 0:
            # done drawing current fractal.
            if self.zoombutton.get_active():
                if self.f.get_param(self.f.MAGNITUDE) > self.minsize:
                    self.select_quadrant_and_zoom()
                else:
                    self.zoombutton.set_active(False)



# A persistent cache of pre-compiled formulas

import os
import stat
import pickle
import hashlib


class TimeStampedObject:
    "An object and the last time it was updated"

    def __init__(self, obj, time, file=None):
        self.time = time
        self.obj = obj
        self.cache_file = file


class T:
    def __init__(self, dir):
        self.dir = os.path.expanduser(dir)
        self.files = {}

    def init(self):
        os.makedirs(self.dir, mode=0o770, exist_ok=True)
        self._loadHash()

    def _indexName(self):
        return os.path.join(self.dir, "index_v00.pkl")

    def _loadHash(self):
        if os.path.exists(self._indexName()):
            self.files = self.loadPickledFile(self._indexName())

    def save(self):
        self.createPickledFile(self._indexName(), self.files)

    def clear(self):
        if not os.path.exists(self.dir):
            return
        for f in os.listdir(self.dir):
            try:
                os.remove(os.path.join(self.dir, f))
            except BaseException:
                pass

    def makefilename(self, name, ext):
        return os.path.join(self.dir, "fract4d_%s%s" % (name, ext))

    def hashcode(self, s, *extras):
        hash = hashlib.md5(s.encode("utf-8"))
        for x in extras:
            hash.update(x.encode("utf-8"))

        return hash.hexdigest()

    def makePickleName(self, s, *extras):
        name = self.hashcode(s, *extras) + ".pkl"
        fullname = os.path.join(self.dir, name)
        return fullname

    def createPickledFile(self, file, contents):
        with open(file, "wb") as f:
            pickle.dump(contents, f, True)
            # print "created %s" % file

    def loadPickledFile(self, file):
        with open(file, "rb") as f:
            contents = pickle.load(f)
        return contents

    def getcontents(self, file, parser):
        mtime = os.stat(file)[stat.ST_MTIME]

        tso = self.files.get(file, None)
        if tso:
            if tso.time == mtime:
                return tso.obj

        with open(file) as f:
            val = parser(f)

        hashname = self.makePickleName(file)
        # self.createPickledFile(hashname,val)
        self.files[file] = TimeStampedObject(val, mtime, hashname)

        return val



#!/usr/bin/env python3
# ------------------------------------------------------------
# fractlexer.py
#
# tokenizer for UltraFractal formula files
# ------------------------------------------------------------
import sys
import re

from . import lex

# set to True to pass through all tokens. This breaks the parser but
# is useful for pretty-printing

keep_all = False

# List of token names.   This is always required
tokens = (
    'NUMBER',
    'COMPLEX',
    'ID',

    'PLUS',
    'MINUS',

    'TIMES',
    'DIVIDE',
    'MOD',

    'LPAREN',
    'RPAREN',
    'LARRAY',
    'RARRAY',
    'MAG',
    'POWER',

    'BOOL_NEG',
    'BOOL_OR',
    'BOOL_AND',

    'EQ',
    'NEQ',
    'LT',
    'LTE',
    'GT',
    'GTE',

    'ASSIGN',

    'COMMENT',
    'NEWLINE',
    'ESCAPED_NL',
    'COMMA',
    'STRING',

    'COMMENT_FORMULA',
    'FORM_ID',
    'FORM_END',
    'SECT_SET',
    'SECT_PARMS',
    'SECT_STM',

    # keywords
    'ELSE',
    'ELSEIF',
    'ENDFUNC',
    'ENDHEADING',
    'ENDIF',
    'ENDPARAM',
    'ENDWHILE',
    'FUNC',
    'HEADING',
    'IF',
    'PARAM',
    'REPEAT',
    'UNTIL',
    'WHILE',

    'TYPE',
    'CONST'
)

# lookup table to convert IDs into keywords
keywords = ["else",
            "elseif",
            "endfunc",
            "endheading",
            "endif",
            "endparam",
            "endwhile",
            "func",
            "heading",
            "if",
            "param",
            "repeat",
            "until",
            "while"]

types = ["bool",
         "color",
         "complex",
         "float",
         "hyper",
         "grad",
         "int",
         "image"]

consts = ["true", "false", "yes", "no"]

lookup = {}
for k in keywords:
    lookup[k] = k.upper()
for t in types:
    lookup[t] = "TYPE"
for c in consts:
    lookup[c] = "CONST"

# Regular expression rules for simple tokens
t_PLUS = r'\+'
t_MINUS = r'-'
t_TIMES = r'\*'
t_DIVIDE = r'/'
t_MOD = r'%'
t_LPAREN = r'\('
t_RPAREN = r'\)'
t_LARRAY = r'\['
t_RARRAY = r'\]'
t_MAG = r'\|'
t_POWER = r'\^'
t_BOOL_NEG = r'!'
t_BOOL_OR = r'\|\|'
t_BOOL_AND = r'&&'
t_EQ = r'=='
t_NEQ = r'!='
t_LT = r'<'
t_LTE = r'<='
t_GT = r'>'
t_GTE = r'>='
t_ASSIGN = r'='
t_COMMA = r','
t_FORM_END = r'\}'

# handle stupid "comment" formula blocks specially
# match ; and Comment because some uf repository files do this


def t_COMMENT_FORMULA(t):
    r';?[Cc]omment\s*{[^}]*}'
    newlines = re.findall(r'\n', t.value)
    t.lexer.lineno += len(newlines)
    pass

# may seem weird, but this includes the starting {
# this is to ensure that the generous pattern match doesn't
# trigger all the time mid-formula (eg, z = "z^2 + c" is a valid formid)


def t_FORM_ID(t):
    r'[^\r\n;"\{]+{'
    # remove trailing whitespace and {
    if not keep_all:
        t.value = re.sub(r"\s*{$", "", t.value)
    return t


def t_NUMBER(t):
    r'(?=\d|\.\d)\d*(\.\d*)?([Ee]([+-]?\d+))?i?'
    if t.value[-1] == "i":  # a complex constant
        t.value = t.value[0:-1]
        t.type = "COMPLEX"

    return t

# these have to be functions to give them higher precedence than ID
# the gnarly regexp syntax is so we are case-insensitive (ick)
# and don't match things like "x = pixel:"

# default, switch, builtin


def t_SECT_SET(t):
    r'(([Dd][Ee][Ff][Aa][Uu][Ll][Tt])|([Ss][Ww][Ii][Tt][Cc][Hh])|([Bb][Uu][Ii][Ll][Tt][Ii][Nn])):'
    if not keep_all:
        t.value = re.sub(":$", "", t.value)
    return t

# a section containing parameter settings, as found in .ugr and .upr files
# gradient, fractal, layer, mapping, formula, inside, outside, alpha, opacity


def t_SECT_PARMS(t):
    r'(([Gg][Rr][Aa][Dd][Ii][Ee][Nn][Tt])|([Ff][Rr][Aa][Cc][Tt][Aa][Ll])|([Ll][Aa][Yy][Ee][Rr])|([Mm][Aa][Pp][Pp][Ii][Nn][Gg])|([Ff][Oo][Rr][Mm][Uu][Ll][Aa])|([Ii][Nn][Ss][Ii][Dd][Ee])|([Oo][Uu][Tt][Ss][Ii][Dd][Ee])|([Aa][Ll][Pp][Hh][Aa])|([Oo][Pp][Aa][Cc][Ii][Tt][Yy])):'
    if not keep_all:
        t.value = re.sub(":$", "", t.value)
    return t

# global, transform, init, loop, final, bailout


def t_SECT_STM(t):
    r'(([Gg][Ll][Oo][Bb][Aa][Ll])|([Tt][Rr][Aa][Nn][Ss][Ff][Oo][Rr][Mm])|([Ii][Nn][Ii][Tt])|([Ll][Oo][Oo][Pp])|([Ff][Ii][Nn][Aa][Ll])|([Bb][Aa][Ii][Ll][Oo][Uu][Tt]))?:'
    if not keep_all:
        t.value = re.sub(":$", "", t.value)
    return t


def t_ID(t):
    r'[@#]?[a-zA-Z_][a-zA-Z0-9_]*'
    global lookup
    lookfor = t.value.lower()  # case insensitive lookup
    if lookfor in lookup:
        t.type = lookup[lookfor]
    return t

# don't produce tokens for newlines preceded by \


def t_ESCAPED_NL(t):
    r'\\\r?\s*\n'
    t.lexer.lineno += 1


def t_COMMENT(t):
    r';[^\n]*'
    if keep_all:
        return t


def t_NEWLINE(t):
    r'\r*\n'
    t.lexer.lineno += 1  # track line numbers
    return t


def t_STRING(t):
    r'"[^"]*"'  # embedded quotes not supported in UF?
    if not keep_all:
        # remove trailing and leading "
        t.value = re.sub(r'(^")|("$)', "", t.value)
        newlines = re.findall(r'\n', t.value)
        t.lexer.lineno += len(newlines)
        t.value = re.sub(r'\\\r?\n[ \t\v]*', "", t.value)  # hide \-split lines
    return t


# A string containing ignored characters (spaces and tabs)
t_ignore = ' \t\r'

# Error handling rule


def t_error(t):
    #print("Illegal character '%s' on line %d" % (t.value[0], t.lexer.lineno))
    t.value = t.value[0]
    t.lexer.skip(1)
    return t


# Build the lexer
lexer = lex.lex(optimize=1)


def get_lexer():
    global lexer
    lexer = lex.lex(optimize=1)
    return lexer


# debugging
if __name__ == '__main__':  # pragma: no cover
    # Test it out
    with open(sys.argv[1], "r") as f:
        data = f.read()

    # Give the lexer some input
    lex.input(data)

    # Tokenize
    while True:
        tok = lex.token()
        if not tok:
            break      # No more input
        print(tok)



# simple types for "assembler" instructions

from .fracttypes import (Bool, Int, Float, Complex, Hyper, Color, Image,
                         TranslationError, typeObjectList)


class ComplexArg:
    ' a pair of args'

    def __init__(self, re, im):
        self.re = re
        self.im = im

    def format(self):
        [self.re.format(), self.im.format()]

    def __str__(self):
        return "Complex(%s,%s)" % (self.re, self.im)


class HyperArg:
    'four args'

    def __init__(self, re, im1, im2, im3):
        self.parts = [re, im1, im2, im3]

    def format(self):
        return [x.format() for x in self.parts]

    def __str__(self):
        return "Hyper(%s,%s,%s,%s)" % tuple(self.parts)


class ColorArg:
    'four args'

    def __init__(self, re, im1, im2, im3):
        self.parts = [re, im1, im2, im3]

    def format(self):
        return [x.format() for x in self.parts]

    def __str__(self):
        return "Color(%s,%s,%s,%s)" % tuple(self.parts)


class ConstArg:
    def __init__(self, value):
        self.value = value


class ConstFloatArg(ConstArg):
    def __init__(self, value):
        ConstArg.__init__(self, value)

    def cformat(self):
        return "%.17f"

    def format(self):
        return "%.17f" % self.value

    def __str__(self):
        return "Float(%s)" % self.format()

    def is_one(self):
        return self.value == 1.0

    def is_zero(self):
        return self.value == 0.0


class ConstIntArg(ConstArg):
    def __init__(self, value):
        ConstArg.__init__(self, value)

    def cformat(self):
        return "%d"

    def format(self):
        return "%d" % self.value

    def __str__(self):
        return "Int(%s)" % self.format()

    def is_one(self):
        return self.value == 1

    def is_zero(self):
        return self.value == 0


class TempArg:
    def __init__(self, value, type):
        self.value = value
        self.type = typeObjectList[type]

    def format(self):
        return self.value

    def cformat(self):
        return self.type.printf

    def __str__(self):
        return "Temp(%s)" % self.format()


def create_arg_from_val(type, val):
    if type == Int or type == Bool or type == Image:
        return ConstIntArg(val)
    elif type == Float:
        return ConstFloatArg(val)
    elif type == Complex:
        return ComplexArg(ConstFloatArg(val[0]), ConstFloatArg(val[1]))
    elif type == Hyper:
        return HyperArg(
            ConstFloatArg(val[0]), ConstFloatArg(val[1]),
            ConstFloatArg(val[2]), ConstFloatArg(val[3]))
    elif type == Color:
        return ColorArg(
            ConstFloatArg(val[0]), ConstFloatArg(val[1]),
            ConstFloatArg(val[2]), ConstFloatArg(val[3]))
    else:
        raise TranslationError(
            "Internal Compiler Error: Unknown constant type %s" % type)


def create_arg(t):
    return create_arg_from_val(t.datatype, t.value)


class Insn:
    'An instruction to be written to output stream'

    def __init__(self, assem):
        self.assem = assem  # string format of instruction

    def source(self):
        return []

    def dest(self):
        return []

    def format(self):
        try:
            lookup = {}
            i = 0
            if self.source() is not None:
                for src in self.source():
                    sname = "s%d" % i
                    lookup[sname] = src.format()
                    i = i + 1
                i = 0

            if self.dest() is not None:
                for dst in self.dest():
                    dname = "d%d" % i
                    lookup[dname] = dst.format()
                    i = i + 1
            return self.assem % lookup
        except Exception as exn:
            print(exn)
            msg = "%s with %s" % (self, lookup)
            raise TranslationError(
                "Internal Compiler Error: can't format " + msg)


class Literal(Insn):
    'A loophole in the system to sneak through text directly'

    def __init__(self, text):
        self.text = text

    def cformat(self):
        return self.text

    def format(self):
        return self.text

    def dest(self):
        return []

    def source(self):
        return []


class Oper(Insn):
    'An operation'

    def __init__(self, assem, src, dst, jumps=[]):
        Insn.__init__(self, assem)
        self.src = src
        self.dst = dst
        self.jumps = jumps

    def dest(self):
        return self.dst

    def source(self):
        return self.src

    def __str__(self):
        return "OPER(%s,[%s],[%s],%s)" % \
               (self.assem,
                ",".join([x.__str__() for x in self.src]),
                ",".join([x.__str__() for x in self.dst]),
                self.jumps)


class Binop(Oper):
    'A binary infix operation, like addition'

    def __init__(self, op, src, dst, generate_trace=False):
        Insn.__init__(self, "")
        self.op = op
        self.src = src
        self.dst = dst
        self.trace = generate_trace

    def const_eval(self):
        c1 = self.src[0]
        c2 = self.src[1]
        klass = c1.__class__
        if self.op == "*":
            val = klass(c1.value * c2.value)
        elif self.op == "+":
            val = klass(c1.value + c2.value)
        elif self.op == "-":
            val = klass(c1.value - c2.value)
        elif self.op == "/":
            val = klass(c1.value / c2.value)
        else:
            # don't know how to const_eval
            return self

        return Move([val], self.dst)

    def __str__(self):
        return "BINOP(%s,[%s],[%s])" % \
               (self.op,
                ",".join([x.__str__() for x in self.src]),
                ",".join([x.__str__() for x in self.dst]))

    def format(self):
        result = "%s = %s %s %s;" % (
            self.dst[0].format(),
            self.src[0].format(),
            self.op,
            self.src[1].format())
        if self.trace:
            result += "printf(\"%s = %s (%s %s %s)\\n\",%s);" % (
                self.dst[0].format(),
                self.dst[0].cformat(),
                self.src[0].format(),
                self.op,
                self.src[1].format(),
                self.dst[0].format())
        return result


class Label(Insn):
    'A label which can be jumped to'

    def __init__(self, label):
        Insn.__init__(self, "%s: ;\n" % label)
        self.label = label

    def format(self):
        return "%s ;\n" % self

    def __str__(self):
        return "%s:" % self.label


class Move(Insn):
    ' A move instruction'

    def __init__(self, src, dst, generate_trace=False):
        Insn.__init__(self, "%(d0)s = %(s0)s;")
        self.src = src
        self.dst = dst
        self.trace = generate_trace

    def dest(self):
        return [self.dst]

    def source(self):
        return [self.src]

    def format(self):
        result = "%s = %s;" % (self.dst[0].format(), self.src[0].format())
        if self.trace:
            result += "printf(\"%s = %s\\n\",%s);" % (
                self.dst[0].format(),
                self.dst[0].cformat(),
                self.src[0].format())
        return result

    def __str__(self):
        return "MOVE(%s,%s,%s)" % (self.assem, self.src, self.dst)


class Decl(Insn):
    ' a variable declaration'

    def __init__(self, assem):
        Insn.__init__(self, assem)
        self.src = None
        self.dst = None

    def __str__(self):
        return "DECL(%s,%s)" % (self.src, self.dst)



import os
import gi

gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, GObject


class DirectorPrefs:
    # wrapper to show dialog for selecting folder
    # returns selected folder or empty string
    def get_folder(self):
        temp_folder = ""
        dialog = Gtk.FileChooserDialog(title="Choose directory...",
                                       transient_for=self.dialog,
                                       action=Gtk.FileChooserAction.SELECT_FOLDER)
        dialog.add_buttons(Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
                           Gtk.STOCK_OPEN, Gtk.ResponseType.OK)
        dialog.set_default_response(Gtk.ResponseType.OK)
        response = dialog.run()
        if response == Gtk.ResponseType.OK:
            temp_folder = dialog.get_filename()
        dialog.destroy()
        return temp_folder

    def create_fct_toggled(self, widget, data=None):
        self.btn_temp_fct.set_sensitive(self.chk_create_fct.get_active())
        self.txt_temp_fct.set_sensitive(self.chk_create_fct.get_active())

    def temp_fct_clicked(self, widget, data=None):
        fold = self.get_folder()
        if fold:
            self.txt_temp_fct.set_text(fold)

    def temp_png_clicked(self, widget, data=None):
        fold = self.get_folder()
        if fold:
            self.txt_temp_png.set_text(fold)

    def __init__(self, animation, parent):
        # pylint: disable=no-member
        self.dialog = Gtk.Dialog(title="Director preferences...",
                                 transient_for=parent,
                                 modal=True, destroy_with_parent=True)
        self.dialog.add_buttons(Gtk.STOCK_OK, Gtk.ResponseType.OK,
                                Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL)

        self.animation = animation
        # -----------Temporary directories---------------------
        #self.frm_dirs=Gtk.Frame("Temporary directories selection")
        # self.frm_dirs.set_border_width(10)
        tbl_dirs = Gtk.Grid()
        tbl_dirs.set_row_spacing(10)
        tbl_dirs.set_column_spacing(10)
        tbl_dirs.set_property("margin", 10)

        lbl_temp_fct = Gtk.Label(label="Temporary directory for .fct files:")
        tbl_dirs.attach(lbl_temp_fct, 0, 1, 1, 1)

        self.txt_temp_fct = Gtk.Entry()
        self.txt_temp_fct.set_text(self.animation.get_fct_dir())
        self.txt_temp_fct.set_sensitive(False)
        tbl_dirs.attach(self.txt_temp_fct, 1, 1, 1, 1)

        self.btn_temp_fct = Gtk.Button(label="Browse")
        self.btn_temp_fct.connect("clicked", self.temp_fct_clicked, None)
        self.btn_temp_fct.set_sensitive(False)
        tbl_dirs.attach(self.btn_temp_fct, 2, 1, 1, 1)

        # this check box goes after (even if it's above above widgets because
        # we connect and change its state here and it change those buttons, so
        # they wouldn't exist
        self.chk_create_fct = Gtk.CheckButton(
            label="Create temporary .fct files")
        self.chk_create_fct.connect("toggled", self.create_fct_toggled, None)
        self.chk_create_fct.set_active(self.animation.get_fct_enabled())
        tbl_dirs.attach(self.chk_create_fct, 0, 0, 1, 1)

        lbl_temp_png = Gtk.Label(label="Temporary directory for .png files:")
        tbl_dirs.attach(lbl_temp_png, 0, 2, 1, 1)

        self.txt_temp_png = Gtk.Entry()
        self.txt_temp_png.set_text(self.animation.get_png_dir())
        tbl_dirs.attach(self.txt_temp_png, 1, 2, 1, 1)

        btn_temp_png = Gtk.Button(label="Browse")
        btn_temp_png.connect("clicked", self.temp_png_clicked, None)
        tbl_dirs.attach(btn_temp_png, 2, 2, 1, 1)

        # self.frm_dirs.add(tbl_dirs)
        self.dialog.vbox.pack_start(tbl_dirs, False, False, 0)
        # self.dialog.vbox.pack_start(self.tbl_main,True,True,0)

        self.dialog.connect('response', self.onResponse)

    # checking is txt fields valid dirs
    def check_fields(self):
        if self.chk_create_fct.get_active():
            # checking fct dir
            if not os.path.isdir(self.txt_temp_fct.get_text()):
                error_dlg = Gtk.MessageDialog(
                    title="Directory for temporary .fct files is not directory",
                    transient_for=self.dialog,
                    modal=True, destroy_with_parent=True)
                error_dlg.add_buttons(
                    Gtk.MessageType.ERROR, Gtk.ButtonsType.OK)
                error_dlg.run()
                error_dlg.destroy()
                return False
        if not os.path.isdir(self.txt_temp_png.get_text()):
            error_dlg = Gtk.MessageDialog(
                title="Directory for temporary .png files is not directory",
                transient_for=self.dialog,
                modal=True, destroy_with_parent=True)
            error_dlg.add_buttons(Gtk.MessageType.ERROR, Gtk.ButtonsType.OK)
            error_dlg.run()
            error_dlg.destroy()
            return False
        return True

    def onResponse(self, widget, id):
        if id == Gtk.ResponseType.OK and not self.check_fields():
            GObject.signal_stop_emission_by_name(self.dialog, "response")

    def show(self):
        self.dialog.show_all()
        response = self.dialog.run()
        if response == Gtk.ResponseType.OK:
            self.animation.set_fct_enabled(self.chk_create_fct.get_active())
            if self.chk_create_fct.get_active():
                self.animation.set_fct_dir(self.txt_temp_fct.get_text())
            self.animation.set_png_dir(self.txt_temp_png.get_text())

        self.dialog.destroy()



import copy

from . import stdlib
from .fracttypes import canBeCast, suffixOfType


class Func:
    def __init__(self, args, ret, fname, pos=-1):
        self.args = args
        self.implicit_args = []
        self.ret = ret
        self.pos = pos
        self.set_func(fname)

    def __copy__(self):
        c = Func(self.args, self.ret, self.fname, self.pos)
        c.implicit_args = copy.copy(self.implicit_args)
        if hasattr(self, "override_name"):
            c.set_override_name(self.override_name)
        if hasattr(self, "caption"):
            c.caption = self.caption
        return c

    def first(self):
        return self

    def set_implicit_arg(self, arg):
        self.implicit_args.append(arg)

    def set_override_name(self, name):
        self.override_name = name

    def set_func(self, fname):
        # compute the name of the stdlib function to call
        # this is similar in purpose to C++ name mangling
        if fname is None:
            self.genFunc = None
        else:
            typed_fname = fname + "_"
            for arg in self.args:
                typed_fname = typed_fname + suffixOfType[arg]
            typed_fname = typed_fname + "_" + suffixOfType[self.ret]

            self.genFunc = stdlib.__dict__.get(typed_fname, typed_fname)

        self.cname = fname
        self.fname = fname

    def matchesArgs(self, potentialArgs):
        if len(potentialArgs) != len(self.args):
            return False
        i = 0
        for arg in self.args:
            if not canBeCast(potentialArgs[i], arg):
                return False
            i = i + 1
        return True



# UI for gathering needed data and storing them in director bean class
# then it calls PNGGeneration, and (if everything was OK) AVIGeneration

# TODO: change default directory when selecting new according to already set item
# (for temp dirs, avi and fct files selections)

import os
import fnmatch
import tempfile
import gi

gi.require_version('Gdk', '3.0')
gi.require_version('Gtk', '3.0')
from gi.repository import Gdk, Gtk, GObject

from fract4d_compiler import fc
from fract4d import animation, fractal, fractconfig
from . import dialog, hig, PNGGen, AVIGen, DlgAdvOpt, director_prefs


class UserCancelledError(Exception):
    pass


class SanityCheckError(Exception):
    "The type of exception which is thrown when animation sanity checks fail"

    def __init__(self, msg):
        Exception.__init__(self, msg)


class DirectorDialog(dialog.T, hig.MessagePopper):
    RESPONSE_RENDER = 1

    def check_for_keyframe_clash(self, keyframe, fct_dir):
        keydir = os.path.dirname(keyframe)
        if keydir[-1] != "/":
            keydir += "/"

        if fct_dir[-1] != "/":
            fct_dir += "/"
        if keydir == fct_dir:
            raise SanityCheckError(
                _("Keyframe %s is in the temporary .fct directory and could be overwritten. Please change temp directory." % keyframe))

    def check_fct_file_sanity(self):
        # things to check with fct temp dir
        if not self.animation.get_fct_enabled():
            # we're not generating .fct files, so this is superfluous
            return

        # check fct temp dir is set
        if self.animation.get_fct_dir() == "":
            raise SanityCheckError(
                _("Directory for temporary .fct files not set"))

        # check if it is directory
        if not os.path.isdir(self.animation.get_fct_dir()):
            raise SanityCheckError(
                _("Path for temporary .fct files is not a directory"))

        fct_path = self.animation.get_fct_dir()

        # heck if any keyframe fct files are in temp fct dir
        for i in range(self.animation.keyframes_count()):
            keyframe = self.animation.get_keyframe_filename(i)
            self.check_for_keyframe_clash(keyframe, fct_path)

        # check if there are any .fct files in temp fct dir
        has_any = False
        for file in os.listdir(fct_path):
            if fnmatch.fnmatch(file, "*.fct"):
                has_any = True

            if has_any is True:
                response = self.ask_question(
                    _("Directory for temporary .fct files contains other .fct files"),
                    _("These may be overwritten. Proceed?"))
                if response != Gtk.ResponseType.ACCEPT:
                    raise UserCancelledError()
            return

    # throws SanityCheckError if there was a problem
    def check_sanity(self):
        # check if at least two keyframes exist
        if self.animation.keyframes_count() < 2:
            raise SanityCheckError(_("There must be at least two keyframes"))

        # check png temp dir is set
        if self.animation.get_png_dir() == "":
            raise SanityCheckError(
                _("Directory for temporary .png files not set"))

        # check if it is directory
        if not os.path.isdir(self.animation.get_png_dir()):
            raise SanityCheckError(
                _("Path for temporary .png files is not a directory"))

        # check avi file is set
        if self.animation.get_avi_file() == "":
            raise SanityCheckError(_("Output AVI file name not set"))

        self.check_fct_file_sanity()

    # wrapper to show dialog for selecting .fct file
    # returns selected file or empty string
    def get_fct_file(self):
        temp_file = ""
        dialog = Gtk.FileChooserDialog(title="Choose keyframe...",
                                       transient_for=self,
                                       action=Gtk.FileChooserAction.OPEN)
        dialog.add_buttons(Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
                           Gtk.STOCK_OPEN, Gtk.ResponseType.OK)
        dialog.set_default_response(Gtk.ResponseType.OK)
        # ----setting filters---------
        filter = Gtk.FileFilter()
        filter.set_name("gnofract4d files (*.fct)")
        filter.add_pattern("*.fct")
        dialog.add_filter(filter)
        filter = Gtk.FileFilter()
        filter.set_name("All files")
        filter.add_pattern("*")
        dialog.add_filter(filter)
        # ----------------------------
        response = dialog.run()
        if response == Gtk.ResponseType.OK:
            temp_file = dialog.get_filename()
        dialog.destroy()
        return temp_file

    # wrapper to show dialog for selecting .avi file
    # returns selected file or empty string
    def get_avi_file(self):
        temp_file = ""
        dialog = Gtk.FileChooserDialog(title="Save AVI file...",
                                       transient_for=self,
                                       action=Gtk.FileChooserAction.SAVE)
        dialog.add_buttons(Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
                           Gtk.STOCK_OPEN, Gtk.ResponseType.OK)
        dialog.set_default_response(Gtk.ResponseType.OK)
        current_file = self.txt_temp_avi.get_text()
        dialog.set_current_name(current_file if current_file else "video.webm")
        # ----setting filters---------
        filter = Gtk.FileFilter()
        filter.set_name("webm video file (*.webm)")
        filter.add_pattern("*.webm")
        dialog.add_filter(filter)
        filter = Gtk.FileFilter()
        filter.set_name("All files")
        filter.add_pattern("*")
        dialog.add_filter(filter)
        # ----------------------------
        response = dialog.run()
        if response == Gtk.ResponseType.OK:
            temp_file = dialog.get_filename()
        dialog.destroy()
        return temp_file

    # wrapper to show dialog for selecting .cfg file
    # returns selected file or empty string
    def get_cfg_file_save(self):
        temp_file = ""
        dialog = Gtk.FileChooserDialog(title="Save animation...",
                                       transient_for=self,
                                       action=Gtk.FileChooserAction.SAVE)
        dialog.add_buttons(Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
                           Gtk.STOCK_OPEN, Gtk.ResponseType.OK)
        dialog.set_default_response(Gtk.ResponseType.OK)
        dialog.set_current_name("animation.fcta")
        # ----setting filters---------
        filter = Gtk.FileFilter()
        filter.set_name("gnofract4d animation files (*.fcta)")
        filter.add_pattern("*.fcta")
        dialog.add_filter(filter)
        filter = Gtk.FileFilter()
        filter.set_name("All files")
        filter.add_pattern("*")
        dialog.add_filter(filter)
        # ----------------------------
        response = dialog.run()
        if response == Gtk.ResponseType.OK:
            temp_file = dialog.get_filename()
        dialog.destroy()
        return temp_file

    # wrapper to show dialog for selecting .fct file
    # returns selected file or empty string
    def get_cfg_file_open(self):
        temp_file = ""
        dialog = Gtk.FileChooserDialog(title="Choose animation...",
                                       transient_for=self,
                                       action=Gtk.FileChooserAction.OPEN)
        dialog.add_buttons(Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
                           Gtk.STOCK_OPEN, Gtk.ResponseType.OK)
        dialog.set_default_response(Gtk.ResponseType.OK)
        # ----setting filters---------
        filter = Gtk.FileFilter()
        filter.set_name("gnofract4d animation files (*.fcta)")
        filter.add_pattern("*.fcta")
        dialog.add_filter(filter)
        filter = Gtk.FileFilter()
        filter.set_name("All files")
        filter.add_pattern("*")
        dialog.add_filter(filter)
        # ----------------------------
        response = dialog.run()
        if response == Gtk.ResponseType.OK:
            temp_file = dialog.get_filename()
        # elif response == Gtk.ResponseType.CANCEL:
        #    print 'Closed, no files selected'
        dialog.destroy()
        return temp_file

    def temp_avi_clicked(self, widget, data=None):
        avi = self.get_avi_file()
        if avi:
            self.txt_temp_avi.set_text(avi)

    def output_width_changed(self, widget, data=None):
        self.animation.set_width(self.spin_width.get_value())

    def output_height_changed(self, widget, data=None):
        self.animation.set_height(self.spin_height.get_value())

    def output_framerate_changed(self, widget, data=None):
        self.animation.set_framerate(self.spin_framerate.get_value())

    def duration_changed(self, widget, data=None):
        if self.current_select == -1:
            return
        self.animation.set_keyframe_duration(
            self.current_select, int(
                self.spin_duration.get_value()))
        self.update_model()

    def stop_changed(self, widget, data=None):
        if self.current_select == -1:
            return
        self.animation.set_keyframe_stop(
            self.current_select, int(
                self.spin_kf_stop.get_value()))
        self.update_model()

    def interpolation_type_changed(self, widget, data=None):
        if self.current_select == -1:
            return
        self.animation.set_keyframe_int(
            self.current_select, int(
                self.cmb_interpolation_type.get_active()))
        self.update_model()

    # point of whole program:)
    # first we generate png files and list, then .avi
    def generate(self, create_avi=True):
        try:
            self.check_sanity()
        except SanityCheckError as exn:
            self.show_error(_("Cannot Generate Animation"), str(exn))
            raise

        png_gen = PNGGen.PNGGeneration(self.animation, self.compiler, self)
        res = png_gen.show()
        if res == 1:
            # user cancelled, but they know that. Stop silently
            return
        elif res != 0:
            # unexpected return code
            return

        if not create_avi:
            return

        avi_gen = AVIGen.AVIGeneration(self.animation, self)
        res = avi_gen.show()
        if res == 1:
            # user cancelled, but they know that. Stop silently
            pass
        elif res == 0:
            # success
            self.show_info(
                _("AVI Generation Complete"),
                _("File is %s." % self.animation.get_avi_file()))

    def generate_clicked(self, widget, data=None):
        self.generate(True)

    def adv_opt_clicked(self, widget, data=None):
        if self.current_select == -1:
            return
        dlg = DlgAdvOpt.DlgAdvOptions(
            self.current_select, self.animation, self)
        dlg.show()

    # before selecting keyframes in list box we must update values of spin
    # boxes in case user typed something in there
    def before_selection(self, selection, data=None, *kwargs):
        self.spin_duration.update()
        self.spin_kf_stop.update()
        return True

    # update right box (duration, stop, interpolation type) when keyframe is
    # selected from list
    def selection_changed(self, widget, data=None):
        (model, it) = self.tv_keyframes.get_selection().get_selected()
        if it is not None:
            # ------getting index of selected row-----------
            index = 0
            it = model.get_iter_first()
            while it is not None:
                if self.tv_keyframes.get_selection().iter_is_selected(it):
                    break
                it = model.iter_next(it)
                index = index + 1
            self.current_select = index
            self.spin_duration.set_value(
                int(self.animation.get_keyframe_duration(index)))
            self.spin_kf_stop.set_value(
                self.animation.get_keyframe_stop(index))
            self.cmb_interpolation_type.set_active(
                self.animation.get_keyframe_int(index))
            return
        else:
            self.spin_duration.set_value(25)
            self.spin_kf_stop.set_value(1)
            self.cmb_interpolation_type.set_active(animation.INT_LINEAR)
            self.current_select = -1

    def update_model(self):
        (model, it) = self.tv_keyframes.get_selection().get_selected()
        if it is not None:
            # ------getting index of selected row-----------
            index = 0
            it = model.get_iter_first()
            while it is not None:
                if self.tv_keyframes.get_selection().iter_is_selected(it):
                    break
                it = model.iter_next(it)
                index = index + 1

            model.set(it, 1, self.animation.get_keyframe_duration(index))
            model.set(it, 2, self.animation.get_keyframe_stop(index))
            int_type = self.animation.get_keyframe_int(index)
            if int_type == animation.INT_LINEAR:
                model.set(it, 3, "Linear")
            elif int_type == animation.INT_LOG:
                model.set(it, 3, "Logarithmic")
            elif int_type == animation.INT_INVLOG:
                model.set(it, 3, "Inverse logarithmic")
            else:
                model.set(it, 3, "Cosine")

    def swap_redblue_clicked(self, widget, data=None):
        self.animation.set_redblue(self.chk_swapRB.get_active())

    def add_from_file(self, widget, data=None):
        file = self.get_fct_file()
        if file != "":
            self.add_keyframe(file)

    def add_from_current(self, widget, data=None):
        (tmp_fd, tmp_name) = tempfile.mkstemp(suffix='.fct')
        with os.fdopen(tmp_fd, 'w') as f:
            self.f.save(f)
        self.add_keyframe(tmp_name)
        return

    def add_keyframe(self, file):
        if file != "":
            # get current seletion
            (model, it) = self.tv_keyframes.get_selection().get_selected()
            if it is None:  # if it's none, just append at the end of the list
                it = model.append([file, 25, 1, "Linear"])
            else:  # append after currently selected
                it = model.insert_after(it, [file, 25, 1, "Linear"])

            # add to bean with default parameters
            if self.current_select != -1:
                self.animation.add_keyframe(
                    file, 25, 1, animation.INT_LINEAR, self.current_select + 1)
            else:
                self.animation.add_keyframe(file, 25, 1, animation.INT_LINEAR)
            # and select newly item
            self.tv_keyframes.get_selection().select_iter(it)
            # set default duration
            self.spin_duration.set_value(25)
            # set default stop
            self.spin_kf_stop.set_value(1)
            # set default interpolation type
            self.cmb_interpolation_type.set_active(animation.INT_LINEAR)

    def add_keyframe_clicked(self, widget, event):
        if event.type == Gdk.EventType.BUTTON_PRESS:
            widget.popup(
                None,
                None,
                None,
                None,
                event.get_button()[1],
                event.time)
            # Tell calling code that we have handled this event the buck
            # stops here.
            return True
        # Tell calling code that we have not handled this event pass it on.
        return False

    def remove_keyframe_clicked(self, widget, data=None):
        # is anything selected
        (model, it) = self.tv_keyframes.get_selection().get_selected()
        if it is not None:
            temp_curr = self.current_select
            model.remove(it)
            self.animation.remove_keyframe(temp_curr)

    def updateGUI(self):
        # keyframes
        (model, it) = self.tv_keyframes.get_selection().get_selected()
        model.clear()
        for i in range(self.animation.keyframes_count() - 1, -1, -1):
            filename = self.animation.get_keyframe_filename(i)
            duration = self.animation.get_keyframe_duration(i)
            stopped = self.animation.get_keyframe_stop(i)
            it = model.insert(0, [filename, duration, stopped, ""])
            int_type = self.animation.get_keyframe_int(i)
            if int_type == animation.INT_LINEAR:
                model.set(it, 3, "Linear")
            elif int_type == animation.INT_LOG:
                model.set(it, 3, "Logarithmic")
            elif int_type == animation.INT_INVLOG:
                model.set(it, 3, "Inverse logarithmic")
            else:
                model.set(it, 3, "Cosine")
        # output part
        self.txt_temp_avi.set_text(self.animation.get_avi_file())
        self.spin_width.set_value(self.animation.get_width())
        self.spin_height.set_value(self.animation.get_height())
        self.spin_framerate.set_value(self.animation.get_framerate())
        self.chk_swapRB.set_active(self.animation.get_redblue())

    # loads configuration file, returns 0 on ok, -1 on error (and displays
    # error message)
    def load_configuration(self, file):
        if file == "":
            return -1
        try:
            self.animation.load_animation(file)
        except Exception as err:
            self.show_error(
                _("Cannot load animation"),
                str(err))
            return -1
        # set GUI to reflect changes
        self.updateGUI()
        return 0

    # loads configuration from pickled file
    def load_configuration_clicked(self, widget, data=None):
        cfg = self.get_cfg_file_open()
        if cfg != "":
            self.load_configuration(cfg)

    # reset all field to defaults
    def new_configuration_clicked(self, widget, data=None):
        self.animation.reset()
        self.updateGUI()

    # save configuration in file
    def save_configuration_clicked(self, widget, data=None):
        cfg = self.get_cfg_file_save()
        if cfg != "":
            try:
                self.animation.save_animation(cfg)
            except Exception as err:
                self.show_error(
                    _("Error saving animation"),
                    str(err))

    def preferences_clicked(self, widget, data=None):
        dlg = director_prefs.DirectorPrefs(self.animation, self)
        dlg.show()

    # creating window...
    def __init__(self, main_window, f, userConfig, conf_file=""):
        #pylint: disable=no-member
        dialog.T.__init__(
            self,
            _("Director"),
            main_window,
            (_("_Render"), DirectorDialog.RESPONSE_RENDER,
             Gtk.STOCK_CLOSE, Gtk.ResponseType.CLOSE)
        )

        hig.MessagePopper.__init__(self)
        self.animation = animation.T(f.compiler, userConfig)
        self.f = f
        self.compiler = f.compiler

        # main VBox
        self.box_main = Gtk.Box.new(Gtk.Orientation.VERTICAL, 0)
        self.box_main.set_homogeneous(False)
        # --------------------menu-------------------------------
        self.manager = Gtk.UIManager()
        accelgroup = self.manager.get_accel_group()
        self.add_accel_group(accelgroup)

        actiongroup = Gtk.ActionGroup.new("Director")
        actiongroup.add_actions([
            ('DirectorMenuAction', None, _('_Director')),
            ('DirectorEditAction', None, _('_Edit')),

            ('DirectorNewAction', Gtk.STOCK_NEW, _('_New Animation'),
             '<control>N', None, self.new_configuration_clicked),
            ('DirectorOpenAction', Gtk.STOCK_OPEN, _('_Open Animation'),
             '<control>O', None, self.load_configuration_clicked),
            ('DirectorSaveAction', Gtk.STOCK_SAVE, _('_Save Animation'),
             '<control>S', None, self.save_configuration_clicked),
            ('DirectorEditPrefsAction', Gtk.STOCK_PREFERENCES, _('_Preferences'),
             '<control>P', None, self.preferences_clicked),
        ])

        self.manager.insert_action_group(actiongroup, 0)

        menu = '''
<ui>
<menubar>
<menu name="Director" action="DirectorMenuAction">
    <menuitem action="DirectorNewAction"/>
    <menuitem action="DirectorOpenAction"/>
    <menuitem action="DirectorSaveAction"/>
</menu>
<menu name="Edit" action="DirectorEditAction">
    <menuitem action="DirectorEditPrefsAction"/>
</menu>
</menubar>
</ui>
'''
        self.manager.add_ui_from_string(menu)

        self.menubar = self.manager.get_widget('/menubar')
        self.box_main.pack_start(self.menubar, False, True, 0)

        # -----------creating popup menu-------------------------------
        # popup menu for keyframes
        self.popup_menu = Gtk.Menu()
        self.mnu_pop_add_file = Gtk.MenuItem.new_with_label("From file")
        self.popup_menu.append(self.mnu_pop_add_file)
        self.mnu_pop_add_file.connect("activate", self.add_from_file, None)
        self.mnu_pop_add_file.show()
        self.mnu_pop_add_current = Gtk.MenuItem.new_with_label(
            "From current fractal")
        self.popup_menu.append(self.mnu_pop_add_current)
        self.mnu_pop_add_current.connect(
            "activate", self.add_from_current, None)
        self.mnu_pop_add_current.show()

        # --------------Keyframes box-----------------------------------
        self.frm_kf = Gtk.Frame.new("Keyframes")
        self.frm_kf.set_border_width(10)
        vbox_kfs = Gtk.Box.new(Gtk.Orientation.VERTICAL, 8)
        button_box_kfs = Gtk.ButtonBox()
        button_box_kfs.set_layout(Gtk.ButtonBoxStyle.SPREAD)
        self.sw = Gtk.ScrolledWindow()
        self.sw.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        self.sw.set_size_request(-1, 100)
#        filenames=Gtk.ListStore(GObject.TYPE_STRING,
#            GObject.TYPE_STRING)
#        self.tv_keyframes=Gtk.TreeView(filenames)
#        self.tv_column_name = Gtk.TreeViewColumn('Keyframes')
#        self.tv_keyframes.append_column(self.tv_column_name)
#        self.tv_column_duration = Gtk.TreeViewColumn('Duration')
#        self.tv_keyframes.append_column(self.tv_column_duration)
#        self.cell_name = Gtk.CellRendererText()
#        self.tv_column_name.pack_start(self.cell_name, True)
#        self.tv_column_name.add_attribute(self.cell_name, 'text', 0)
#        self.cell_duration = Gtk.CellRendererText()
#        self.tv_column_name.pack_start(self.cell_duration, True)
#        self.tv_column_name.add_attribute(self.cell_duration, 'text', 0)
        filenames = Gtk.ListStore(GObject.TYPE_STRING, GObject.TYPE_UINT,
                                  GObject.TYPE_UINT, GObject.TYPE_STRING)
        self.tv_keyframes = Gtk.TreeView()
        self.tv_keyframes.set_model(filenames)
        column = Gtk.TreeViewColumn(
            'Keyframes', Gtk.CellRendererText(), text=0)
        self.tv_keyframes.append_column(column)

        column = Gtk.TreeViewColumn(
            'Stopped for', Gtk.CellRendererText(), text=2)
        self.tv_keyframes.append_column(column)

        column = Gtk.TreeViewColumn(
            'Transition duration',
            Gtk.CellRendererText(),
            text=1)
        self.tv_keyframes.append_column(column)

        column = Gtk.TreeViewColumn(
            'Interpolation type',
            Gtk.CellRendererText(),
            text=3)
        self.tv_keyframes.append_column(column)

        self.sw.add(self.tv_keyframes)
        self.tv_keyframes.get_selection().connect(
            "changed", self.selection_changed, None)
        self.tv_keyframes.get_selection().set_select_function(self.before_selection, None)
        self.current_select = -1

        self.btn_add_keyframe = Gtk.Button.new_with_label("Add")
        # self.btn_add_keyframe.connect("clicked",self.add_keyframe_clicked,None)
        self.btn_add_keyframe.connect_object(
            "event", self.add_keyframe_clicked, self.popup_menu)

        self.btn_remove_keyframe = Gtk.Button.new_with_label("Remove")
        self.btn_remove_keyframe.connect(
            "clicked", self.remove_keyframe_clicked, None)

        button_box_kfs.pack_start(self.btn_add_keyframe, True, True, 0)
        button_box_kfs.pack_start(self.btn_remove_keyframe, True, True, 0)

        vbox_kfs.pack_start(self.sw, True, True, 0)
        vbox_kfs.pack_start(button_box_kfs, True, True, 10)

        self.frm_kf.add(vbox_kfs)
        self.box_main.pack_start(self.frm_kf, True, True, 0)

        # current keyframe box
        self.current_kf = Gtk.Frame.new("Current Keyframe")
        self.current_kf.set_border_width(10)

        self.box_main.pack_start(self.current_kf, True, True, 0)

        self.tbl_keyframes_right = Gtk.Grid()
        self.tbl_keyframes_right.set_column_homogeneous(True)
        self.tbl_keyframes_right.set_row_homogeneous(True)
        self.tbl_keyframes_right.set_row_spacing(10)
        self.tbl_keyframes_right.set_column_spacing(10)
        self.tbl_keyframes_right.set_border_width(10)

        self.lbl_kf_stop = Gtk.Label(label="Keyframe stopped for:")
        self.tbl_keyframes_right.attach(self.lbl_kf_stop, 0, 0, 1, 1)

        adj_kf_stop = Gtk.Adjustment.new(1, 1, 10000, 1, 10, 0)
        self.spin_kf_stop = Gtk.SpinButton()
        self.spin_kf_stop.set_adjustment(adj_kf_stop)
        self.spin_kf_stop.connect("output", self.stop_changed, None)
        self.tbl_keyframes_right.attach(self.spin_kf_stop, 1, 0, 1, 1)

        self.lbl_duration = Gtk.Label(label="Transition duration:")
        self.tbl_keyframes_right.attach(self.lbl_duration, 0, 1, 1, 1)

        adj_duration = Gtk.Adjustment.new(25, 1, 10000, 1, 10, 0)
        self.spin_duration = Gtk.SpinButton()
        self.spin_duration.set_adjustment(adj_duration)
        self.spin_duration.connect("output", self.duration_changed, None)
        self.tbl_keyframes_right.attach(self.spin_duration, 1, 1, 1, 1)

        self.lbl_int_type = Gtk.Label(label="Interpolation type:")
        self.tbl_keyframes_right.attach(self.lbl_int_type, 0, 2, 1, 1)

        self.cmb_interpolation_type = Gtk.ComboBoxText()  # Gtk.ComboBox()
        self.cmb_interpolation_type.append_text("Linear")
        self.cmb_interpolation_type.append_text("Logarithmic")
        self.cmb_interpolation_type.append_text("Inverse logarithmic")
        self.cmb_interpolation_type.append_text("Cosine")
        self.cmb_interpolation_type.set_active(0)
        self.cmb_interpolation_type.connect(
            "changed", self.interpolation_type_changed, None)
        self.tbl_keyframes_right.attach(
            self.cmb_interpolation_type, 1, 2, 1, 1)

        self.btn_adv_opt = Gtk.Button(label="Advanced options")
        self.btn_adv_opt.connect("clicked", self.adv_opt_clicked, None)
        self.tbl_keyframes_right.attach(self.btn_adv_opt, 0, 3, 2, 1)

        self.current_kf.add(self.tbl_keyframes_right)
        # -------------------------------------------------------------------
        # ----------------------output box-----------------------------------
        self.frm_output = Gtk.Frame.new("Output options")
        self.frm_output.set_border_width(10)

        self.box_output_main = Gtk.Box(orientation=Gtk.Orientation.VERTICAL,
                                       homogeneous=True, spacing=10)
        self.box_output_main.set_border_width(10)
        self.box_output_file = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL,
                                       homogeneous=False, spacing=10)

        self.lbl_temp_avi = Gtk.Label(label="Resulting video file:")
        self.box_output_file.pack_start(self.lbl_temp_avi, False, False, 10)

        self.txt_temp_avi = Gtk.Entry()
        self.box_output_file.pack_start(self.txt_temp_avi, True, True, 10)

        self.btn_temp_avi = Gtk.Button(label="Browse")
        self.btn_temp_avi.connect("clicked", self.temp_avi_clicked, None)
        self.box_output_file.pack_start(self.btn_temp_avi, True, True, 10)

        self.box_output_main.pack_start(self.box_output_file, True, True, 0)

        self.box_output_res = Gtk.HBox(homogeneous=False, spacing=10)

        self.lbl_res = Gtk.Label(label="Resolution:")
        self.box_output_res.pack_start(self.lbl_res, False, False, 10)

        adj_width = Gtk.Adjustment.new(640, 320, 2048, 10, 100, 0)
        self.spin_width = Gtk.SpinButton()
        self.spin_width.set_adjustment(adj_width)
        self.box_output_res.pack_start(self.spin_width, False, False, 10)

        self.lbl_x = Gtk.Label(label="x")
        self.box_output_res.pack_start(self.lbl_x, False, False, 10)

        adj_height = Gtk.Adjustment.new(480, 240, 1536, 10, 100, 0)
        self.spin_height = Gtk.SpinButton()
        self.spin_height.set_adjustment(adj_height)
        self.box_output_res.pack_start(self.spin_height, False, False, 10)

        self.box_output_main.pack_start(self.box_output_res, True, True, 0)

        self.box_output_framerate = Gtk.HBox(homogeneous=False, spacing=10)

        self.lbl_framerate = Gtk.Label(label="Frame rate:")
        self.box_output_framerate.pack_start(
            self.lbl_framerate, False, False, 10)

        adj_framerate = Gtk.Adjustment.new(25, 5, 100, 1, 5, 0)
        self.spin_framerate = Gtk.SpinButton()
        self.spin_framerate.set_adjustment(adj_framerate)
        self.box_output_framerate.pack_start(
            self.spin_framerate, False, False, 10)

        self.chk_swapRB = Gtk.CheckButton(label="Swap red and blue component")
        self.box_output_framerate.pack_start(self.chk_swapRB, False, False, 50)

        self.box_output_main.pack_start(
            self.box_output_framerate, True, True, 0)

        self.frm_output.add(self.box_output_main)
        self.box_main.pack_start(self.frm_output, False, False, 0)

        # check if video converter can be found
        self.converterpath = fractconfig.T.find_on_path("ffmpeg")
        if not self.converterpath:
            # put a message at the bottom to warn user
            warning_box = Gtk.HBox()
            image = Gtk.Image.new_from_icon_name(
                "dialog-warning", Gtk.IconSize.BUTTON)

            warning_box.pack_start(image, True, True, 0)

            message = Gtk.Label(label=_(
                "ffmpeg utility not found. Without it we can't generate any video but can still save sequences of still images."))

            message.set_line_wrap(True)
            warning_box.pack_start(message, True, True, 0)
            self.box_main.pack_end(warning_box, True, True, 0)

        # initialise default settings
        if conf_file:
            self.load_configuration(conf_file)
        else:
            self.updateGUI()

        # don't connect signals until after settings initialised
        self.spin_height.connect(
            "value-changed",
            self.output_height_changed,
            None)
        self.spin_width.connect(
            "value-changed",
            self.output_width_changed,
            None)
        self.spin_framerate.connect(
            "value-changed", self.output_framerate_changed, None)
        self.chk_swapRB.connect("toggled", self.swap_redblue_clicked, None)

        # --------------showing all-------------------------------
        self.vbox.add(self.box_main)
        self.vbox.show_all()

    def onResponse(self, widget, id):
        if id == Gtk.ResponseType.CLOSE or \
                id == Gtk.ResponseType.NONE or \
                id == Gtk.ResponseType.DELETE_EVENT:
            self.hide()
        elif id == DirectorDialog.RESPONSE_RENDER:
            self.animation.set_avi_file(self.txt_temp_avi.get_text())
            try:
                self.generate(self.converterpath is not None)
            except (SanityCheckError, UserCancelledError):
                # prevent dialog closing if being run
                GObject.signal_stop_emission_by_name(self, "response")
            else:
                self.hide()

    def main(self):
        Gtk.main()


if __name__ == "__main__":
    GObject.threads_init()
    Gtk.threads_init()

    main_window = Gtk.Window()
    userConfig = fractconfig.userConfig()
    compiler = fc.Compiler(userConfig)
    f = fractal.T(compiler)
    fracwin = DirectorDialog(main_window, f, userConfig)

    fracwin.main()



# fract compiler datatypes

# order is significant - we use X > Y on types
Bool = 0
Int = 1
Float = 2
Complex = 3
Color = 4
String = 5
Hyper = 6
Gradient = 7
Image = 8

VoidArray = 9
IntArray = 10
FloatArray = 11
ComplexArray = 12


class Type(object):
    def __init__(self, **kwds):
        self.suffix = kwds["suffix"]
        self.printf = kwds.get("printf")  # optional
        self.typename = kwds["typename"]
        self.default = kwds["default"]
        self.slots = kwds.get("slots", 1)
        self.cname = kwds["cname"]
        self.typeid = kwds["id"]
        self.part_names = kwds.get("parts", [""])

    def init_val(self, var):
        return "0"


class FloatType(Type):
    def __init__(self, **kwds):
        Type.__init__(self, **kwds)

    def init_val(self, var):
        if var.param_slot == -1:
            return ["%.17f" % var.value]
        else:
            return ["t__pfo->p[%d].doubleval" % var.param_slot]


class ComplexType(Type):
    def __init__(self, **kwds):
        Type.__init__(self, **kwds)

    def init_val(self, var):
        if var.param_slot == -1:
            return [
                "%.17f" % var.value[0],
                "%.17f" % var.value[1]]
        else:
            return [
                "t__pfo->p[%d].doubleval" % var.param_slot,
                "t__pfo->p[%d].doubleval" % (var.param_slot + 1)]


class QuadType(Type):
    def __init__(self, **kwds):
        Type.__init__(self, **kwds)

    def init_val(self, var):
        if var.param_slot == -1:
            return ["%.17f" % x for x in var.value]
        else:
            return [
                "t__pfo->p[%d].doubleval" %
                x for x in range(var.param_slot, var.param_slot + 4)]


class IntType(Type):
    def __init__(self, **kwds):
        Type.__init__(self, **kwds)

    def init_val(self, var):
        if var.param_slot == -1:
            return ["%d" % var.value]
        else:
            return [
                "t__pfo->p[%d].intval" % var.param_slot]


class GradientType(Type):
    def __init__(self, **kwds):
        Type.__init__(self, **kwds)

    def init_val(self, var):
        if var.param_slot == -1:
            raise TranslationError(
                "Internal Compiler Error: gradient not initialized as a param")
        else:
            return ["t__pfo->p[%d].gradient" % var.param_slot]

        return []


class ImageType(Type):
    def __init__(self, **kwds):
        Type.__init__(self, **kwds)

    def init_val(self, var):
        if var.param_slot == -1:
            raise TranslationError(
                "Internal Compiler Error: image not initialized as a param")
        else:
            return ["t__pfo->p[%d].image" % var.param_slot]

        return []


# these have to be in the indexes given by the constants above
typeObjectList = [
    IntType(id=Bool, suffix="b", printf="%d", typename="bool",
            default=0, cname="int"),

    IntType(id=Int, suffix="i", printf="%d", typename="int",
            default=0, cname="int"),

    FloatType(id=Float, suffix="f", printf="%g", typename="float",
              default=0.0, cname="double"),

    ComplexType(id=Complex, suffix="c", typename="complex",
                default=[0.0, 0.0], slots=2, cname="double", parts=["_re", "_im"]),

    QuadType(id=Color, suffix="C", typename="color",
             default=[0.0, 0.0, 0.0, 0.0], slots=4, cname="double",
             parts=["_re", "_i", "_j", "_k"]),

    Type(id=String, suffix="S", typename="string",
         default="", slots=0, cname="<Error>"),

    QuadType(id=Hyper, suffix="h", typename="hyper",
             default=[0.0, 0.0, 0.0, 0.0], slots=4, cname="double",
             parts=["_re", "_i", "_j", "_k"]),

    GradientType(id=Gradient, suffix="G", typename="gradient",
                 default=0, cname="void *"),

    ImageType(id=Image, suffix="I", typename="image",
              default=0, cname="void *"),

    Type(id=VoidArray, suffix="av", typename="voidarray",
         default=0, cname="void *"),

    Type(id=IntArray, suffix="ai", typename="intarray",
         default=0, cname="int *"),

    Type(id=FloatArray, suffix="af", typename="floatarray",
         default=0, cname="double *"),

    Type(id=ComplexArray, suffix="ac", typename="complexarray",
         default=0, cname="double *"),

]

typeList = [
    Bool,
    Int,
    Float,
    Complex,
    Color,
    String,
    Hyper,
    Gradient,
    Image,
    VoidArray,
    IntArray,
    FloatArray,
    ComplexArray]

suffixOfType = {
    Int: "i",
    Float: "f",
    Complex: "c",
    Bool: "b",
    Color: "C",
    String: "S",
    Hyper: "h",
    Gradient: "G",
    Image: "I",
    VoidArray: "av",
    IntArray: "ai",
    FloatArray: "af",
    ComplexArray: "ac"
}

_typeOfStr = {
    "int": Int,
    "float": Float,
    "complex": Complex,
    "bool": Bool,
    "color": Color,
    "string": String,
    "hyper": Hyper,
    "grad": Gradient,
    "image": Image,
    "array": VoidArray,
    "int array": IntArray,
    "float array": FloatArray,
    "complex array": ComplexArray
}

_strOfType = {
    Int: "int",
    Float: "float",
    Complex: "complex",
    Bool: "bool",
    Color: "color",
    None: "none",
    String: "string",
    Hyper: "hyper",
    Gradient: "grad",
    Image: "image",
    VoidArray: "array",
    IntArray: "int array",
    FloatArray: "float array",
    ComplexArray: "complex array"
}

_slotsForType = {
    Int: 1,
    Float: 1,
    Complex: 2,
    Bool: 1,
    Color: 4,
    String: 0,
    Hyper: 4,
    Gradient: 1,
    Image: 1
    # arrays not supported as args
}


def elementSizeOf(type):
    if type == IntArray:
        size = 4
    elif type == FloatArray:
        size = 8
    elif type == ComplexArray:
        size = 8
    else:
        raise TranslationError("invalid allocation type %s" % type)

    return size


def arrayTypeOf(t, node=None):
    if t == Int:
        return IntArray
    if t == Float:
        return FloatArray
    if t == Complex:
        return ComplexArray

    # there is no array type, make a friendly error
    if not node:
        pos = ""
    else:
        pos = "%d: " % node.pos
    raise TranslationError(
        "%sArrays of type %s are not supported" % (pos, strOfType(t)))


def elementTypeOf(t, node=None):
    if t == IntArray:
        return Int
    if t == FloatArray:
        return Float
    if t == ComplexArray:
        return Complex

    # there is no array type, make a friendly error
    if not node:
        pos = ""
    else:
        pos = "%d: " % node.pos
    raise TranslationError(
        "%sArrays of type %s are not supported" % (pos, strOfType(t)))


def typeOfStr(tname):
    if not tname:
        return None
    return _typeOfStr[tname.lower()]


def strOfType(t):
    return _strOfType[t]


def default_value(t):
    return typeObjectList[t].default


def slotsForType(t):
    return typeObjectList[t].slots


_canBeCast = [
    # rows are from, columns are to
    # Bool Int Float Complex Color String Hyper Gradient Image VoidArray,
    # IntArray FloatArray ComplexArray
    [1, 1, 1, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, ],  # Bool
    [1, 1, 1, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, ],  # Int
    [1, 0, 1, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, ],  # Float
    [1, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, ],  # Complex
    [0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, ],  # Color
    [0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, ],  # String
    [1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, ],  # Hyper
    [0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, ],  # Gradient
    [0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, ],  # Image
    [0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, ],  # VoidArray
    [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, ],  # IntArray
    [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, ],  # FloatArray
    [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, ]  # ComplexArray
]


def canBeCast(t1, t2):
    ' can t1 be cast to t2?'
    if t1 is None or t2 is None:
        return 0
    return _canBeCast[t1][t2]

# a convenient place to put this.


class TranslationError(Exception):
    def __init__(self, msg):
        Exception.__init__(self, msg)
        self.msg = msg


class InternalCompilerError(TranslationError):
    def __init__(self, msg):
        TranslationError.__init__(self, "Internal Compiler Error:" + msg)


class Var:
    def __init__(self, type, value=None, pos=-1, **kwds):
        #assert(type_ != None)
        # assert(isinstance(pos,types.IntType))
        self._set_type(type)
        if value is None:
            self.value = self._typeobj.default
        else:
            self.value = value
        self.pos = pos
        self.cname = None
        self.__doc__ = kwds.get("doc")
        self.declared = False
        self.param_slot = -1

    def _get_is_temp(self):
        return False
    is_temp = property(_get_is_temp)

    def _get_type(self):
        return self._typeobj.typeid

    def _set_type(self, t):
        self._typeobj = typeObjectList[t]

    type = property(_get_type, _set_type)

    def _get_ctype(self):
        return self._typeobj.cname

    ctype = property(_get_ctype)

    def struct_name(self):
        return "t__pfo->" + self.cname

    def first(self):
        return self

    def __str__(self):
        return "%s %s (%d)" % (strOfType(self.type), self.value, self.pos)

    def init_val(self):
        try:
            return self._typeobj.init_val(self)
        except AttributeError:
            print("%s type for %s" % (self._typeobj.typename, self.cname))
            raise

    def _get_part_names(self):
        return self._typeobj.part_names

    part_names = property(_get_part_names)


class Temp(Var):
    def __init__(self, type_, name):
        Var.__init__(self, type_)
        self.cname = name

    def _get_is_temp(self):
        return True

    is_temp = property(_get_is_temp)




from gi.repository import GObject


class HistoryEntry:
    def __init__(self, redo, redo_data, undo, undo_data):
        self.undo_action = undo
        self.undo_data = undo_data
        self.redo_action = redo
        self.redo_data = redo_data

    def undo(self):
        self.undo_action(self.undo_data)

    def redo(self):
        self.redo_action(self.redo_data)


class Sequence(GObject.GObject):
    __gsignals__ = {
        'can-undo': (
            GObject.SignalFlags.RUN_FIRST, None, (GObject.TYPE_BOOLEAN,)),
        'can-redo': (
            GObject.SignalFlags.RUN_FIRST, None, (GObject.TYPE_BOOLEAN,))
    }

    def __init__(self):
        GObject.GObject.__init__(self)
        self.pos = 0  # the position after the current item
        self.history = []

    def can_undo(self):
        return self.pos > 0

    def can_redo(self):
        return self.pos < len(self.history)

    def send_signals(self):
        self.emit('can-undo', self.can_undo())
        self.emit('can-redo', self.can_redo())

    def do(self, redo_action, redo_data, undo_action, undo_data):
        # replace everything from here on with the new item
        if self.pos < len(self.history):
            undo_action = self.history[self.pos].undo_action
            undo_data = self.history[self.pos].undo_data

        del self.history[self.pos:]
        self.history.append(
            HistoryEntry(redo_action, redo_data, undo_action, undo_data))
        self.pos = len(self.history)

        self.send_signals()

    def undo(self):
        if not self.can_undo():
            raise ValueError("Can't Undo at start of sequence")
        self.pos -= 1
        self.history[self.pos].undo()
        self.send_signals()

    def redo(self):
        if not self.can_redo():
            raise ValueError("Can't Redo at end of sequence")
        self.history[self.pos].redo()
        self.pos += 1

        self.send_signals()

    def make_undo_sensitive(self, widget):
        # make this widget only be sensitive if we can undo

        def set_sensitivity(sequence, can_undo):
            widget.set_sensitive(can_undo)

        self.connect('can-undo', set_sensitivity)
        self.emit('can-undo', self.can_undo())

    def make_redo_sensitive(self, widget):
        # make this widget only be sensitive if we can undo

        def set_sensitivity(sequence, can_redo):
            widget.set_sensitive(can_redo)

        self.connect('can-redo', set_sensitivity)
        self.emit('can-redo', self.can_redo())


GObject.type_register(Sequence)



#!/usr/bin/env python3

# create HTML document documenting the standard library
# this gets included into the main manual

import sys
from xml.sax.saxutils import escape

from fract4d_compiler import fracttypes, fsymbol


def strOfType(t):
    return fracttypes.strOfType(t).capitalize()


class SymbolPrinter:
    def __init__(self, f):
        self.f = f
        self.funcs = {}
        self.operators = {}
        self.vars = {}

    def add_symbol(self, key, val):
        if key.startswith("@_") or key.startswith("_"):
            # skip internal symbols
            return
        if isinstance(val, fracttypes.Var):
            self.vars[key] = val
        elif isinstance(val, fsymbol.OverloadList):
            if val.is_operator():
                self.operators[key] = val
            else:
                self.funcs[key] = val

    def output_entry(self, nrows=1):
        print(
            '<td valign="top" align="left" rowspan="%d">' %
            nrows, file=self.f)

    def output_refentry_header(self, key, val, type, nrows=1):
        print('<tr>', file=self.f)
        self.output_entry(nrows)
        print('%s</td>' % escape(key), file=self.f)

    def output_overload(self, func):
        self.output_entry()
        print(", ".join(map(strOfType, func.args)), file=self.f)
        print('</td>', file=self.f)
        self.output_entry()
        print('%s</td>' % strOfType(func.ret), file=self.f)

    def output_function(self, val):
        self.output_overload(val[0])

        for func in val[1:]:
            print('</tr>', file=self.f)
            print('<tr>', file=self.f)
            self.output_overload(func)

    def output_refentry_footer(self):
        print('</tr>', file=self.f)
        #print >>self.f, '<row><entry>&nbsp;</entry></row>'

    def output_refentry_body(self, val, nrows=1):
        self.output_entry(nrows)
        text = val.__doc__ or "No documentation yet."
        print(escape(text), file=self.f)
        print('</td>', file=self.f)

    def output_symbol(self, key, val, type):
        if isinstance(val, fsymbol.OverloadList):
            nrows = len(val)
            self.output_refentry_header(key, val, type, nrows)
            self.output_refentry_body(val, nrows)
            self.output_function(val)
        else:
            self.output_refentry_header(key, val, type)
            self.output_refentry_body(val)
            print('<td>%s</td>' % strOfType(val.type), file=self.f)

        self.output_refentry_footer()

    def output_all(self):
        self.output_table(self.operators, "Operators", "operator")
        self.output_table(self.funcs, "Functions", "function")
        self.output_table(self.vars, "Symbols", "(symbol)")

    def output_table(self, table, name, type):
        print('<h2>%s</h2>' % escape(name), file=self.f)
        print('<table>', file=self.f)
        #print('<colgroup cols="4"/>', file=self.f)
        print('''
<thead>
<tr>
                    <th>Name</th>
                    <th>Description</th>
                    <th>Argument Types</th>
                    <th>Return Type</th>
</tr>
</thead>''', file=self.f)
        print('<tbody>', file=self.f)

        keys = list(table.keys())
        keys.sort()
        for k in keys:
            self.output_symbol(k, table[k], type)
        print('</tbody>', file=self.f)
        print('</table>', file=self.f)

def main(outfile):
    with open(outfile, "w") as out:
        # insert front matter
        print('''---
title: "Standard Library Reference"
draft: false
---
''', file=out)
        d = fsymbol.T()
        printer = SymbolPrinter(out)

        for k in list(d.default_dict.keys()):
            printer.add_symbol(d.demangle(k), d[k])

        printer.output_all()


if __name__ == '__main__':
    main(sys.argv[1])
